import re
from collections import Counter
from typing import Any, Collection, Iterable, Iterator, Mapping, Optional, Pattern, Tuple, TypeVar, Union

from yellowbox.extras.webserver.util import MismatchReason, reason_is_ne

T = TypeVar('T')

_missing = object()
K = TypeVar('K')
V = TypeVar('V')


def _is_submap_of(submap: Mapping[K, V], supermap: Mapping[K, V]):
    """
    checks whether a map is a submap of another map
    Args:
        submap: the submap to check
        supermap: the supermap to check

    Returns:
        True if the condition exists, or a MismatchReason otherwise.

    Notes:
        Mapping A is a submap of Mapping B iff every pair in A also exists in B
    """
    for k, v in submap.items():
        other_v = supermap.get(k, _missing)  # type:ignore[arg-type]
        if other_v is _missing:
            return MismatchReason(f'expected key {k}, none found')
        if other_v != v:
            return MismatchReason(f'expected key {k} to have value {v}, got {other_v}')
    return True


def _is_submultimap_of(submultimap: Mapping[K, Collection[V]], supermultimap: Mapping[K, Collection[V]]):
    """
    checks whether a map is a submultimap of another map
    Args:
        submultimap: the submultimap to check
        supermultimap: the supermultimap to check

    Returns:
        True if the condition exists, or a MismatchReason otherwise.

    Notes:
        Mapping A is a submap of Mapping B iff for every key k in A, Counter(B[k]) >= Counter(A[k])
    """
    for k, v in submultimap.items():
        other_v = supermultimap.get(k, _missing)
        if other_v is _missing:
            return MismatchReason(f'expected key {k}, none found')
        if not isinstance(other_v, Iterable):
            return MismatchReason(f'expected key {k} to be in iterable {other_v}')
        diff = Counter(other_v)
        diff.subtract(Counter(v))
        missing_value = next((k for (k, v) in diff.items() if v < 0), _missing)
        if missing_value is not _missing:
            expected = sum(i == missing_value for i in v)
            return MismatchReason(f'expected value {missing_value} to appear {expected} times in key {k}'
                                  f', got {other_v}')
    return True


def _to_expected_map(name: str, exact: Optional[T], submap: Optional[T]) -> Optional[Tuple[T, bool]]:
    """
    A utility function to convert two parameters one representing an exact-match expectation and one a submap
     expectation, into a single value.

    Args:
        name: the name of the exact match parameter
        exact: the exact-match parameter value
        submap: the subset-match parameter value

    Returns:
        a 2-tuple, containing the expected value, and whether the value is an expected submap, or None if no parameters
         were provided
    """
    if exact is not None:
        if submap is not None:
            raise TypeError(f'at most one of "{name}" and "{name}_submap" can be provided')
        return exact, False
    if submap is not None:
        return submap, True
    return None


def _matches_expected_map(name: str, expected: Optional[Tuple[Mapping[K, V], bool]], recorded: Mapping[K, V]) \
        -> Optional[str]:
    """
    Matches a map against the return value of _to_expected_map
    Args:
        name: the name of the field being matched
        expected: the expected value, as returned by _to_expected_map
        recorded: the recorded value to match

    Returns:
        True if the condition matches, or a MismatchReason otherwise.
    """
    if expected is None:
        return None
    expected_value, is_subset = expected
    if is_subset:
        submap_match = _is_submap_of(expected_value, recorded)
        if not submap_match:
            return f'{name} mismatch: {submap_match}'
    else:
        if expected_value != recorded:
            return reason_is_ne(name, expected_value, recorded)
    return None


def _matches_expected_multimap(name: str, expected: Optional[Tuple[Mapping[K, Collection[V]], bool]],
                               recorded: Mapping[K, Collection[V]]) -> Optional[str]:
    """
    Matches a multimap against the return value of _to_expected_map
    Args:
        name: the name of the field being matched
        expected: the expected value, as returned by _to_expected_map
        recorded: the recorded value to match

    Returns:
        None if the condition matches, or a string describing the reason otherwise.
    """
    if expected is None:
        return None
    expected_value, is_subset = expected
    if is_subset:
        submap_match = _is_submultimap_of(expected_value, recorded)
        if not submap_match:
            return f'{name} mismatch: {submap_match}'
    else:
        if expected_value != recorded:
            return reason_is_ne(name, expected_value, recorded)
    return None


def _repr_map(name: str, expected: Optional[Tuple[T, bool]]):
    """
    A utility function to convert the return value of _to_expected_map back to the provided parameters needed to
     construct it.
    Args:
        name: the name of the exact-match parameter
        expected: the return value of _to_expected_map for these parameters

    Returns:
        a mapping of the parameters that would result in this expectation if passed to the constructor
    """
    if expected is None:
        return {}
    expected_value, is_subset = expected
    if is_subset:
        return {f'{name}_submap': expected_value}
    else:
        return {name: expected_value}


class ScopeExpectation:
    """
    A utility mixin class representing an expected http scope, common to both HTTP and websockets.
    """

    def __init__(self, headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                 headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                 path: Optional[Union[str, Pattern[str]]] = None, path_params: Optional[Mapping[str, Any]] = None,
                 path_params_submap: Optional[Mapping[str, Any]] = None,
                 query_params: Optional[Mapping[str, Collection[str]]] = None,
                 query_params_submap: Optional[Mapping[str, Collection[str]]] = None, ):
        self.headers = _to_expected_map('headers', headers, headers_submap)
        if isinstance(path, str):
            self.path_pattern = re.compile(re.escape(path))
        else:
            self.path_pattern = path
        self.path_params = _to_expected_map('path_params', path_params, path_params_submap)
        self.query_params = _to_expected_map('query_params', query_params, query_params_submap)

    def scope_mismatch_reasons(self, recorded) -> Iterator[str]:
        header_match = _matches_expected_multimap('header', self.headers, recorded.headers)
        if header_match:
            yield header_match

        if (self.path_pattern is not None
                and self.path_pattern.fullmatch(str(recorded.path)) is None):
            yield reason_is_ne('path', self.path_pattern.pattern, recorded.path)

        path_params_match = _matches_expected_map('path_params', self.path_params, recorded.path_params)
        if path_params_match:
            yield path_params_match

        query_params_match = _matches_expected_multimap('query_params', self.query_params, recorded.query_params)
        if query_params_match:
            yield query_params_match

    def _repr_map(self):
        """
        Returns: a mapping of the parameters used to create this mixin instance
        """
        args = {
            **_repr_map('headers', self.headers),
            **_repr_map('path_params', self.path_params),
            **_repr_map('query_params', self.query_params),
        }
        if self.path_pattern is not None:
            args['path'] = self.path_pattern
        return args



from __future__ import annotations

from contextlib import contextmanager
from functools import partial
from typing import (
    TYPE_CHECKING, Awaitable, Callable, ContextManager, Iterable, Iterator, List, Optional, Tuple, Union, overload
)

from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.routing import BaseRoute, Route
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR, WS_1011_INTERNAL_ERROR
from starlette.websockets import WebSocket

from yellowbox.extras.webserver.http_request_capture import RecordedHTTPRequest, RecordedHTTPRequests
from yellowbox.extras.webserver.ws_request_capture import RecordedWSTranscripts, recorder_websocket_endpoint

HTTP_SIDE_EFFECT = Union[Response, Callable[[Request], Awaitable[Response]]]
WS_SIDE_EFFECT = Callable[[WebSocket], Awaitable[Optional[int]]]
METHODS = Union[str, Iterable[str]]

if TYPE_CHECKING:
    from yellowbox.extras.webserver.webserver import WebServer


class EndpointPatch(ContextManager):
    """
    A the return value of endpoint side effect patches, restores the original side effect if ever exited
    """

    def __init__(self, endpoint: Union[MockHTTPEndpoint, MockWSEndpoint], restore_side_effect):
        self.endpoint = endpoint
        self.restore_side_effect = restore_side_effect

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.endpoint.side_effect = self.restore_side_effect
        self.endpoint.owner._raise_from_pending()


class MockHTTPEndpoint:
    """
    A mock http endpoint for a webserver
    """

    def __init__(self, name: str, methods: METHODS, rule_string: str, side_effect: HTTP_SIDE_EFFECT,
                 auto_read_body: bool = True, forbid_implicit_head_verb: bool = True):
        """
        Args:
            name: the name of the endpoint
            methods: the methods by which the endpoint is accessible
            rule_string: the rule-string of the endpoint, as specified by starlette routes
            side_effect: the side effect of the endpoint. This can be either a starlette response, or an async callable
             that accepts a starlette request and returns a starlette response.
            auto_read_body: If true (the default), the request is always awaited to be received in full before replying
             to the client.
            forbid_implicit_head_verb: By default, starlette automatically makes all routes that are accessible by the
             GET method also accessible through the HEAD method. If this parameter is set to true (the default),
             this behaviour is disabled and HEAD must be added explicitly if required.
        """
        self._request_captures: List[RecordedHTTPRequests] = []
        self.owner: Optional[WebServer] = None

        if isinstance(methods, str):
            methods = (methods,)

        self.methods: Tuple[str, ...] = tuple(m.upper() for m in methods)
        self.rule_string = rule_string
        self.__name__ = name
        self.side_effect = side_effect
        self.auto_read_body = auto_read_body
        self.forbid_implicit_head_verb = forbid_implicit_head_verb

    def patch(self, side_effect: HTTP_SIDE_EFFECT):
        """
        Change the side effect of the endpoint
        Args:
            side_effect: the new side effect of the endpoint

        Returns:
            A context manager that, if exited, restores the endpoint's original side effect
        """
        self.owner._raise_from_pending()
        previous_side_effect = self.side_effect
        self.side_effect = side_effect
        return EndpointPatch(self, previous_side_effect)

    @contextmanager
    def capture_calls(self) -> Iterator[RecordedHTTPRequests]:
        """
        A context manager that records all requests to the endpoint in its scope to a list
        Returns:
             A RecordedHTTPRequests to which the requests are recorded
        """
        self.owner._raise_from_pending()
        if not self.auto_read_body:
            raise RuntimeError("cannot capture calls if auto_read_body is disabled")
        calls = RecordedHTTPRequests()
        self._request_captures.append(calls)
        try:
            yield calls
        finally:
            if self._request_captures[-1] is not calls:
                raise RuntimeError('capture_calls contexts cannot be used in parallel')
            self._request_captures.pop()
            self.owner._raise_from_pending()

    async def get(self, request: Request):
        # the target of the starlette route
        if not self.owner:
            raise RuntimeError('endpoint must be assigned to a webserver')
        if self.owner._pending_exception:
            return PlainTextResponse(
                f'an exception in the webserver had previously occurred: {self.owner._pending_exception!r}',
                status_code=HTTP_500_INTERNAL_SERVER_ERROR
            )
        try:
            if isinstance(self.side_effect, Response):
                ret = self.side_effect
            else:
                ret = await self.side_effect(request)
        except Exception as ex:
            self.owner._pending_exception = ex
            return PlainTextResponse(f'handler raised an exception: {ex!r}', status_code=HTTP_500_INTERNAL_SERVER_ERROR)
        else:
            if self.auto_read_body:
                await request.body()
            if self._request_captures:  # recording the request reads its body, which we might not want
                recorded = await RecordedHTTPRequest.from_request(request)
                for c in self._request_captures:
                    c.append(recorded)
        return ret

    def route(self) -> BaseRoute:
        """
        Returns: a starlette route representing the endpoint
        """
        ret = Route(self.rule_string, self.get, methods=self.methods, name=self.__name__)  # type:ignore[arg-type]
        if self.forbid_implicit_head_verb and 'HEAD' not in self.methods:
            ret.methods.discard('HEAD')
        return ret


@overload
def http_endpoint(methods: METHODS, rule_string: str, *, auto_read_body: bool = True, forbid_head_verb: bool = True) \
        -> Callable[[HTTP_SIDE_EFFECT], MockHTTPEndpoint]: ...


@overload
def http_endpoint(methods: METHODS, rule_string: str, side_effect: HTTP_SIDE_EFFECT, *, auto_read_body: bool = True,
                  forbid_implicit_head_verb: bool = True) -> MockHTTPEndpoint: ...


def http_endpoint(methods: METHODS, rule_string: str, side_effect: Optional[HTTP_SIDE_EFFECT] = None, **kwargs):
    """
    Create a mock HTTP endpoint.
    Args:
        methods: forwarded to MockHTTPEndpoint
        rule_string: forwarded to MockHTTPEndpoint
        side_effect: forwarded to MockHTTPEndpoint
        **kwargs: forwarded to MockHTTPEndpoint

    Returns:
        an http endpoint

    Notes:
        can also be used as a decorator, when decorating the side effect

    Examples:
        >>> @http_endpoint('GET', '/bar')
        ... async def bar(request: Request):
        ...   return PlainTextResponse(int(request.query_params['x'])**2)
        ...
        ... WebServer().add_http_endpoint(bar)

    """

    def ret(func: HTTP_SIDE_EFFECT):
        try:
            name = func.__name__  # type: ignore[union-attr]
        except AttributeError:
            name = f'{methods} {rule_string}'
        return MockHTTPEndpoint(name, methods, rule_string, func, **kwargs)

    if side_effect is None:
        return ret
    return ret(side_effect)


class MockWSEndpoint:
    """
    A mock websocket endpoint for a webserver
    """

    def __init__(self, name: str, rule_string: str, side_effect: WS_SIDE_EFFECT):
        """
        Args:
            name: the name of the endpoint
            rule_string: the rule-string of the endpoint, as specified by starlette routes
            side_effect: the side effect of the endpoint. This should be an async callable that accepts a starlette
             websocket and optionally returns an int to close the connection with that code.
        """
        self._request_captures: List[RecordedWSTranscripts] = []
        self.owner: Optional[WebServer] = None

        self.rule_string = rule_string
        self.__name__ = name
        self.side_effect = side_effect

        # this will be the endpoint handed to the starlette.WebSocketRoute.
        # the route chooses whether to hand off the scope, receive, send functions (which we need) only if we fail
        # inspect.isfunction and inspect.ismethod. Fortunately, partial fails these tests. By storing it as an
        # attribute of the endpoint, we will be able to better locate routes relating to the endpoint
        self.endpoint = partial(recorder_websocket_endpoint, function=self.get, sinks=self._request_captures)

    async def get(self, websocket: WebSocket):
        # this will be the function that our endpoint will eventually route to
        if not self.owner:
            await websocket.close(WS_1011_INTERNAL_ERROR)  # note that this will actually send a 403 code :shrug:
            return
        if self.owner._pending_exception:
            await websocket.close(WS_1011_INTERNAL_ERROR)  # note that this will actually send a 403 code :shrug:
            return
        try:
            code = await self.side_effect(websocket)
        except Exception as ex:
            self.owner._pending_exception = ex
            await websocket.close(WS_1011_INTERNAL_ERROR)
        else:
            if code is not None:
                await websocket.close(code)

    def patch(self, side_effect: WS_SIDE_EFFECT):
        """
        Change the side effect of the endpoint
        Args:
            side_effect: the new side effect of the endpoint

        Returns:
            A context manager that, if exited, restores the endpoint's original side effect
        """
        self.owner._raise_from_pending()
        previous_side_effect = self.side_effect
        self.side_effect = side_effect
        return EndpointPatch(self, previous_side_effect)

    @contextmanager
    def capture_calls(self) -> Iterator[RecordedWSTranscripts]:
        self.owner._raise_from_pending()
        calls = RecordedWSTranscripts()
        self._request_captures.append(calls)
        try:
            yield calls
        finally:
            if self._request_captures[-1] is not calls:
                raise RuntimeError('capture_calls contexts cannot be used in parallel')
            self._request_captures.pop()
            self.owner._raise_from_pending()


@overload
def ws_endpoint(rule_string: str) -> Callable[[WS_SIDE_EFFECT], MockWSEndpoint]: pass


@overload
def ws_endpoint(rule_string: str, side_effect: WS_SIDE_EFFECT) -> MockWSEndpoint: pass


def ws_endpoint(rule_string: str, side_effect: Optional[WS_SIDE_EFFECT] = None):
    """
    Create a mock websocket endpoint.
    Args:
        rule_string: forwarded to MockWSEndpoint
        side_effect: forwarded to MockWSEndpoint

    Returns:
        a websocket endpoint

    Notes:
        can also be used as a decorator, when decorating the side effect

    Examples:
        >>> @ws_endpoint('/square')
        ... async def square(ws: WebSocket):
        ...   await ws.accept()
        ...   x = int(await ws.receive_text())
        ...   await ws.send_text(str(x*x))
        ...   await ws.close()
        ...
        ... WebServer().add_ws_endpoint(square)
    """

    def ret(func: WS_SIDE_EFFECT):
        return MockWSEndpoint(func.__name__, rule_string, func)

    if side_effect is None:
        return ret
    return ret(side_effect)



from abc import ABC, abstractmethod
from contextlib import contextmanager, nullcontext
from typing import ContextManager, Iterator, Optional, Sequence, Type, TypeVar

from docker import DockerClient
from docker.models.containers import Container
from docker.models.networks import Network

import yellowbox.networks as networks_mod
from yellowbox.containers import _DEFAULT_TIMEOUT, get_aliases, is_alive, is_removed
from yellowbox.retry import RetrySpec
from yellowbox.service import YellowService
from yellowbox.utils import _get_spinner


class ContainerService(YellowService):
    def __init__(self, containers: Sequence[Container], remove=True):
        """
        Notes:
             The containers should be ordered so that it is safe to start them in order, and to re
        """
        self.containers = containers
        self.remove = remove

    @abstractmethod
    def start(self, retry_spec: Optional[RetrySpec] = None):
        """
        Start the service. Wait for startup by repeatedly attempting an operation until success

        Args:
            retry_spec: The specifications for the repeated attempts. If not provided,
             a predefined default RetrySpec should be used.

        Returns:
            self, for usage as a context manager.

        Notes:
            This method should block until the service is fully operational.

        """
        for c in self.containers:
            if c.status.lower() == 'started':
                continue
            c.start()
            c.reload()
        return super(ContainerService, self).start()

    def stop(self, signal='SIGTERM'):
        for c in reversed(self.containers):
            if not is_alive(c):
                continue
            c.kill(signal)
            c.wait(timeout=_DEFAULT_TIMEOUT)
            c.reload()
            if self.remove:
                c.remove(v=True)

    def is_alive(self):
        return all(is_alive(c) for c in self.containers)

    @property
    def _endpoint_containers(self) -> Sequence[Container]:
        return self.containers

    def connect(self, network: Network):
        """
        Add the service to an external docker network.
        """
        for ec in self._endpoint_containers:
            network.connect(ec)
            ec.reload()

    def disconnect(self, network: Network, **kwargs):
        """
        Remove the service from an external docker network.

        Note: Implementors should take care not to disconnect removed containers.
        """
        for ec in reversed(self._endpoint_containers):
            if not is_removed(ec):
                network.disconnect(ec, **kwargs)
                ec.reload()


class SingleEndpointService(ContainerService):
    @property
    @abstractmethod
    def _single_endpoint(self) -> Container:
        pass

    def _endpoint_containers(self) -> Sequence[Container]:
        return self._single_endpoint,

    def connect(self, network: Network, **kwargs) -> Sequence[str]:
        network.connect(self._single_endpoint, **kwargs)
        self._single_endpoint.reload()
        return get_aliases(self._single_endpoint, network)

    def disconnect(self, network: Network, **kwargs):
        if not is_removed(self._single_endpoint):
            network.disconnect(self._single_endpoint, **kwargs)
            self._single_endpoint.reload()


class SingleContainerService(SingleEndpointService, ABC):
    def __init__(self, container: Container, **kwargs):
        super().__init__((container,), **kwargs)

    @property
    def container(self):
        return self.containers[0]

    @property
    def _single_endpoint(self) -> Container:
        return self.container


_T = TypeVar("_T", bound=ContainerService)


class RunMixin:
    @classmethod
    def service_name(cls):
        return cls.__name__

    @classmethod
    @contextmanager
    def run(cls: Type[_T], docker_client: DockerClient, *, spinner: bool = True,  # type: ignore
            retry_spec: Optional[RetrySpec] = None,
            network: Optional[Network] = None, **kwargs) -> Iterator[_T]:
        """
        Same as RunMixin.run, but allows to forward retry arguments to the blocking start method.

        Args:
            docker_client: a DockerClient instance to use when creating the service
            spinner: whether or not to use a yaspin spinner
            retry_spec: forwarded to cls.start
            network: connect service to network
            **kwargs: all keyword arguments are forwarded to the class's constructor
        """
        yaspin_spinner = _get_spinner(spinner)
        with yaspin_spinner(f"Fetching {cls.service_name()} ..."):  # type: ignore[attr-defined]
            service = cls(docker_client, **kwargs)

        connect_network: ContextManager[None]
        if network:
            connect_network = networks_mod.connect(network, service)
        else:
            connect_network = nullcontext()

        with connect_network:
            with yaspin_spinner(f"Waiting for {cls.service_name()} to start..."):  # type: ignore[attr-defined]
                service.start(retry_spec=retry_spec)
            with service:
                yield service



from yellowbox._version import __version__
from yellowbox.clients import docker_client
from yellowbox.image_build import build_image
from yellowbox.networks import connect, temp_network
from yellowbox.retry import RetrySpec
from yellowbox.service import YellowService
from yellowbox.subclasses import ContainerService, RunMixin, SingleContainerService, SingleEndpointService

__all__ = [
    '__version__',
    'docker_client',
    'build_image',
    'connect', 'temp_network',
    'RetrySpec',
    'YellowService',
    'ContainerService', 'RunMixin', 'SingleContainerService', 'SingleEndpointService',

]



from __future__ import annotations

import logging
from contextlib import contextmanager

try:
    from functools import cached_property  # type: ignore
# Support 3.7
except ImportError:
    cached_property = property  # type: ignore

import re
import threading
from functools import partial, wraps
from threading import RLock
from typing import (
    Any, Callable, Dict, Generator, Iterable, Iterator, List, Optional, Pattern, TypeVar, Union, no_type_check
)
from urllib.parse import urlparse
from weakref import WeakMethod

from simple_websocket_server import WebSocket, WebSocketServer

from yellowbox.service import YellowService
from yellowbox.utils import docker_host_name

logger = logging.getLogger(__name__)


class _WebsocketTemplate(WebSocket):
    """Template websocket protocol.

    Used in WebSocketServer, this class handles incoming connections and
    communicates over a generator to send and receive data. See WebsocketService
    for usage.
    """
    _get_generator: Optional[WeakMethod] = None
    """
    A weakmethod to call WebsocketService._get_generator and find the
    generator appropriate for the connected websocket path.
    """
    _generator: "Optional[_GENERAOTR_TYPE]" = None
    """An initialized generator for IO with the websocket"""

    def connected(self) -> None:
        """Websocket connected callback.

        Finds the correct generator for IO with this websocket and sets it
        up.
        """
        initialized = False
        try:
            path = urlparse(self.request.path).path
            if path is None:
                # Not supposed to happen.
                logger.warning("Message received with no path")
                return

            assert self._get_generator
            get_generator = self._get_generator()  # Resolve weakref
            assert get_generator
            generator_function = get_generator(path)

            if generator_function is None:
                logger.info(f"No handler assigned to {path}. "
                            "Closing connection.")
                return

            self._generator = generator_function(self)
            initialized = True
        except Exception:
            logger.exception("Exception raised on generator start.")
            raise

        finally:
            if not initialized:
                self.close()

        self._advance_generator(None)

    def _advance_generator(
            self, data: Union[bytearray, str, None]) -> None:
        """Advance the IO generator. Send it data and wait for output."""
        try:
            try:
                # First one will send None, rest will send data.
                msg = self._generator.send(data)  # type: ignore
            except StopIteration as exc:
                if exc.value:  # type: ignore
                    # Sending may throw an exception.
                    self.send_message(exc.value)  # type: ignore
                self.close()
            else:
                if msg is not None:
                    self.send_message(msg)
        except Exception:
            logger.exception("Exception raised while handling callback.")
            raise  # Handle_close will be called

    def handle(self) -> None:
        """Incoming data callback."""
        self._advance_generator(self.data)

    def handle_close(self) -> None:
        """Connection closed callback."""
        err = ConnectionAbortedError()
        generator = self._generator

        # Closed before full initialization
        if generator is None:
            return
        del self._generator

        try:
            generator.throw(err)
        except Exception as exc:
            # Check if it's our exception. If it is, ignore it.
            if exc is err:
                return
            logger.exception("Exception raised on connection close.")
            raise
        finally:
            generator.close()  # May throw an exception. Ignore it.


@no_type_check
def _to_generator(
        side_effect: SIDE_EFFECT_TYPE) -> _GEN_FUNCTION_TYPE:
    """Convert a side effect to an IO generator function.

    Args:
        side_effect: See WebsocketService.route().

    Returns:
        Generator function, same as the one in WebsocketService.route().
    """
    # Side effect == normal string
    if isinstance(side_effect, (str, bytes, bytearray, memoryview)):
        def gen(*args: Any, **kwargs: Any) -> _GENERAOTR_TYPE:
            return side_effect
            yield  # On purpose.
        return gen

    # Side effect == list of strings
    if isinstance(side_effect, Iterable):
        def gen(*args, **kwargs) -> _GENERAOTR_TYPE:
            for item in side_effect:
                yield item
        return gen

    assert callable(side_effect)

    @wraps(side_effect)  # type: ignore # Mypy GH-10002
    def gen(*args: Any, **kwargs: Any) -> _GENERAOTR_TYPE:
        # Side effect == normal function that returns a string.
        result = side_effect(*args, **kwargs)
        if result is None or isinstance(result, (str, bytes,
                                                 bytearray, memoryview)):
            return result

        # Side effect == generator function
        return (yield from result)

    return gen


# Type aliases used all around
_YIELDTYPES = Union[str, bytes, bytearray, memoryview, None]

_GENERAOTR_TYPE = Generator[_YIELDTYPES,
                            Union[bytearray, str], _YIELDTYPES]

_GEN_FUNCTION_TYPE = Callable[[WebSocket], _GENERAOTR_TYPE]

SIDE_EFFECT_TYPE = Union[
    _YIELDTYPES, List[_YIELDTYPES],
    Callable[[WebSocket], Optional[_YIELDTYPES]],
    _GEN_FUNCTION_TYPE
]

_T = TypeVar("_T")


class WebsocketService(YellowService):
    """Yellobox service to handle incoming websocket connections.

    Allows 2 sided communication using different kinds of side effects.
    Non-blocking, all side effects are ran in a different thread.

    Example:
        >>> service = WebsocketService()

        Let's write "Hello!" to the websocket upon connection to the "/hello"
        endpoint:
        >>> service.add("Hello!", "/hello")

        We'll also make a simple echo endpoint:
        >>> @service.route("/echo")
        ... def echo(websocket):
        ...     data = None
        ...     while True:
        ...         # Yield sends out the data, and waits for incoming data.
        ...         data = yield data
        ...

        And start our service!
        >>> service.start()

        Keep in mind it's non-blocking. The service will run in the background.

        An local websocket should connect to `service.local_url` or
        `service.container_url` if communicating with a hosted docker
        container.
    """
    port: int
    """Server listening port."""

    def __init__(self) -> None:
        """Initialize the service."""
        self._routes: Dict[str, Callable] = {}
        self._re_routes: Dict[Pattern[str], Callable] = {}
        self._lock = RLock()
        _stop_event = self._stop_event = threading.Event()

        class _WebsocketImpl(_WebsocketTemplate):
            _get_generator = WeakMethod(self._get_generator)  # type: ignore

        server = self._server = WebSocketServer("0.0.0.0", 0, _WebsocketImpl)
        self.port = server.serversocket.getsockname()[1]

        # serve_forever() doesn't let you close the server without throwing
        # an exception >.<
        def loop() -> None:
            while not _stop_event.is_set():
                server.handle_request()

        self._thread = threading.Thread(target=loop, daemon=True)

    @cached_property
    def local_url(self) -> str:
        """Local URL for connecting on the same host.

        Suffix this with the path. For example, if you wish to connect locally
        to the "/echo" endpoint, connect to `service.local_url + '/echo'`.
        """
        return f'ws://127.0.0.1:{self.port}'

    @cached_property
    def container_url(self) -> str:
        """URL for connecting over a locally hosted docker container.

        Suffix this with the path. For example, if you wish to connect to the
        "/echo" endpoint from inside a container, connect
        to `service.container_url + '/echo'`.
        """
        return f'ws://{docker_host_name}:{self.port}'

    def is_alive(self) -> bool:
        """Boolean stating if wesocket service is active."""
        return self._thread.is_alive()

    # Mypy doesn't support self generics yet without manual binding (bound=)
    # https://mypy.readthedocs.io/en/stable/generics.html#generic-methods-and-generic-self
    @no_type_check
    def start(self: _T) -> _T:
        """Start the service.

        Non-blocking. Service runs in the background.
        """
        self._thread.start()
        return super().start()

    def stop(self) -> None:
        """Stop the service."""
        self._stop_event.set()
        self._thread.join(1)

        # Shouldn't actually happen.
        if self._thread.is_alive():
            logger.error("Failed to stop the service gracefully. Timed out.")

        self._server.close()
        return super().stop()

    def _get_generator(self, path: str) -> Union[_GEN_FUNCTION_TYPE, None]:
        """Get the IO generator function appropriate to the connected path.

        Args:
            path: Websocket path.

        Returns:
            Generator function, or None if not found.
        """
        callback = self._routes.get(path)
        if callback:
            return callback

        items = reversed(tuple(self._re_routes.items()))  # Thread safety
        for pattern, callback in items:
            if pattern.fullmatch(path):
                return callback

        return None

    def route(self, path: Optional[str] = None, *,
              regex: Optional[Union[Pattern[str], str]] = None
              ) -> Callable[[SIDE_EFFECT_TYPE], None]:
        """Add a route using a decorator syntax.

        Raises an exception if the route already exists.

        Example:

            >>> @service.route("/echo")
            ... def echo(websocket):
            ...     data = None
            ...     while True:
            ...         # Yield sends out the data, and waits for incoming data.
            ...         data = yield data
            ...

        Args:
            path: path to accept connections on. Omit if using regex.
            regex: path regex to accept connections on. Omit if using path.

        Returns:
            Decorator for adding a side effect as a route.
        """
        if path and regex:
            raise ValueError("Only one of path or regex can be specified.")

        if not (path or regex):
            raise ValueError("path or regex must be specified.")

        return partial(self.add, path=path, regex=regex)

    def add(self, side_effect: SIDE_EFFECT_TYPE,
            path: Optional[str] = None, *,
            regex: Optional[Union[Pattern[str], str]] = None,
            _overwrite: bool = False) -> None:
        """Add a route.

        Raises an exception if the route already exists.

        Args:
            side_effect: Side effect can be either a string, bytes or bytearray
            to return a single message; iterable of any combination of them to
            return multiple messages; A regular function that returns any of them
            that will be called upon receiving an incoming connection; or a
            generator function that yields any of them and accepts data. For
            more information see the examples above.
            path: path to accept connections on. Omit if using regex.
            regex: path regex to accept connections on. Omit if using path.

        Raises:
            RuntimeError: Trying to add a route while it already exists.
        """
        if path and regex:
            raise ValueError("Only one of path or regex can be specified.")

        if not (path or regex):
            raise ValueError("path or regex must be specified.")

        if regex:
            regex = re.compile(regex)
        elif not isinstance(path, str):
            # Prevent variable order confusion.
            raise TypeError("path must be a string.")

        gen = _to_generator(side_effect)

        # Check and place
        with self._lock:
            if not _overwrite and (path in self._routes or
                                   regex in self._re_routes):
                raise RuntimeError(f"Route {path or regex} already exists!")

            if path:
                self._routes[path] = gen
            else:
                self._re_routes[regex] = gen  # type: ignore

    @contextmanager
    def patch(self, side_effect: SIDE_EFFECT_TYPE,
              path: Optional[str] = None, *,
              regex: Optional[Union[Pattern[str], str]] = None) -> Iterator[None]:
        """Temporarily patch a route.

        Like `.add()` but uses a context manager for easy removal.
        """
        self.add(side_effect, path, regex=regex)
        try:
            yield
        finally:
            self.remove(path, regex=regex)

    def set(self, side_effect: SIDE_EFFECT_TYPE,
            path: Optional[str] = None, *,
            regex: Optional[Union[Pattern[str], str]] = None) -> None:
        """Set a route.

        Like `add()` but overwrites existing routes.
        """
        self.add(side_effect=side_effect, path=path,
                 regex=regex, _overwrite=True)

    def remove(self, path: Optional[str] = None, *,
               regex: Optional[Union[Pattern[str], str]] = None) -> None:
        """Remove a route.

        Args:
            path: Path that was previously inserted. Omit if using regex.
            regex: Regex that was previously inserted. Omit if using path.
        """
        if path and regex:
            raise ValueError("Only one of path or regex can be specified.")

        if not (path or regex):
            raise ValueError("path or regex must be specified.")

        if path:
            del self._routes[path]
        else:
            assert regex  # For Mypy
            regex = re.compile(regex)
            with self._lock:
                del self._re_routes[regex]

    def clear(self) -> None:
        """Remove all routes."""
        with self._lock:
            self._routes.clear()
            self._re_routes.clear()



from __future__ import annotations

from dataclasses import dataclass
from json import loads as json_loads
from typing import Any, Callable, Collection, Dict, List, Mapping, Optional, Pattern, Sequence, Tuple, Union, overload

from igraph import Graph
from starlette.requests import Request

from yellowbox.extras.webserver.request_capture import ScopeExpectation
from yellowbox.extras.webserver.util import MismatchReason, reason_is_ne

_missing = object()


class ExpectedHTTPRequest(ScopeExpectation):
    """
    An expected HTTP request
    """

    def __init__(self, headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                 headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                 path: Optional[Union[str, Pattern[str]]] = None, path_params: Optional[Mapping[str, Any]] = None,
                 path_params_submap: Optional[Mapping[str, Any]] = None,
                 query_params: Optional[Mapping[str, Collection[str]]] = None,
                 query_params_submap: Optional[Mapping[str, Collection[str]]] = None, method: Optional[str] = None,
                 body: Optional[bytes] = None, text: Optional[str] = None, json: Any = _missing,
                 content_predicate: Optional[Union[Callable[[bytes], bool], Tuple[Callable[[bytes], Any], Any]]]
                 = None):
        """
        Args:
            headers: If specified, expects the request to have these headers exactly
            headers_submap: If specified expects the request to have at least the headers specified
            path: If specified, expected the request url (after the host) to either be equal to this (in case of string)
             or to fully match the regex pattern provided
            path_params: If specified, expects the request to have exactly these path parameters, as provided by the
             starlette request
            path_params_submap: if specified, expects the result to have at least theses path parameters, as provided
             by the starlette request
            query_params: If specified, expects the request to have these query arguments exactly
            query_params_submap: If specified, expects the request to have at least these query arguments
            method: If specified, expects the request to have that HTTP method (case-insensitive)
            body: If specified, expects the request to have that exact byte content. Cannot be used alongside other
             content-testing parameters
            text: If specified, expects the request to have that exact text content (using strict utf-8 decoding).
             Cannot be used alongside other content-testing parameters
            json: If specified, expects the JSON-decoded content of the request to be equal to the specified value.
             Cannot be used alongside other content-testing parameters
            content_predicate: If specified, must be either a callable that accepts a bytes object, or a tuple of a
             callable that accepts a bytes object and another value, and expects the return value of the callable with
             the content of the request to evaluate to either True or the second element of the tuple, if one is
             provided. Cannot be used alongside other content-testing parameters
        """
        super().__init__(headers, headers_submap, path, path_params, path_params_submap, query_params,
                         query_params_submap)

        if method is None:
            self.method = None
        else:
            self.method = method.upper()

        if (
                (body is not None)
                + (text is not None)
                + (json is not _missing)
                + (content_predicate is not None)
        ) >= 2:
            raise ValueError('only one of content, text, json must be set')

        self.body_decode: Optional[Callable[[bytes], Any]]
        self.data: Optional[Any]
        if body is not None:
            self.body_decode = lambda x: x
            self.data = body
        elif text is not None:
            self.body_decode = lambda x: x.decode()
            self.data = text
        elif json is not _missing:
            self.body_decode = lambda x: json_loads(x.decode())
            self.data = json
        elif isinstance(content_predicate, tuple):
            self.body_decode, self.data = content_predicate
        elif callable(content_predicate):
            self.body_decode = content_predicate
            self.data = True
        else:
            self.body_decode = self.data = None

    def matches(self, recorded: RecordedHTTPRequest) -> Union[bool, MismatchReason]:
        """
        Test if an http request meets the expectations of self
        Args:
            recorded: a recorded http request

        Returns:
            True if the request matches, or a MismatchReason object with the reason why otherwise.
        """
        reasons = list(self.scope_mismatch_reasons(recorded))

        if self.method and self.method != recorded.method:
            reasons.append(reason_is_ne('body', self.method, recorded.method))

        if self.body_decode is not None:
            try:
                body = self.body_decode(recorded.content)
            except Exception as e:
                reasons.append(f'failed to parse content: {e!r}')
            else:
                if self.data != body:
                    reasons.append(reason_is_ne('content', self.data, body))

        if reasons:
            return MismatchReason(', '.join(reasons))
        return True

    def __repr__(self):
        args = self._repr_map()
        if self.method is not None:
            args['method'] = self.method
        if self.body_decode is not None:
            args['content'] = self.data

        return 'ExpectedHTTPRequest(' + ', '.join(f'{k}={v!r}' for (k, v) in args.items()) + ')'


@dataclass
class RecordedHTTPRequest:
    """
    A recorded HTTP request, received by a starlette application.
    """
    headers: Mapping[str, Sequence[str]]
    method: str
    path: str
    path_params: Mapping[str, Any]
    query_params: Mapping[str, Sequence[str]]
    content: bytes

    @classmethod
    async def from_request(cls, request: Request):
        """
        Create a new recorded request from a starlette request.
        Args:
            request: the active starlette request

        Returns:
            the recorded recorded request

        Notes:
            this method waits for and reads the request body
        """
        headers = {}
        for k, v in request.headers.items():
            if k not in headers:
                headers[k] = [v]
            else:
                headers[k].append(v)

        query_args = {}
        for k, v in request.query_params.multi_items():
            if k not in query_args:
                query_args[k] = [v]
            else:
                query_args[k].append(v)

        return cls(
            headers,
            request.method,
            request.url.path,
            request.path_params,
            query_args,
            await request.body()
        )


class RecordedHTTPRequests(List[RecordedHTTPRequest]):
    """
    A list of recorded HTTP requests, in the order they were received
    """

    def assert_not_requested(self):
        """
        asserts that no requests were recorded.
        """
        if self:
            raise AssertionError(f'{len(self)} requests, latest: {self[-1]}')

    def assert_requested(self):
        """
        asserts that at least one request was recorded.
        """
        if not self:
            raise AssertionError('No requests were made')

    def assert_requested_once(self):
        """
        asserts that exactly one request was recorded.
        """
        if not self:
            raise AssertionError('No requests were made')
        if len(self) > 1:
            raise AssertionError('Multiple requests were made:'
                                 + ''.join(f'\n\t{existing}' for existing in self))

    @overload
    def assert_requested_with(self, expected: ExpectedHTTPRequest):
        ...

    @overload
    def assert_requested_with(self, *, headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                              headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                              path: Optional[Union[str, Pattern[str]]] = None,
                              path_params: Optional[Mapping[str, Any]] = None,
                              path_params_submap: Optional[Mapping[str, Any]] = None,
                              query_params: Optional[Mapping[str, Collection[str]]] = None,
                              query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                              method: Optional[str] = None, body: Optional[bytes] = None,
                              text: Optional[str] = None, json: Any = _missing,
                              content_predicate:
                              Optional[Union[Callable[[bytes], bool], Tuple[Callable[[bytes], Any], Any]]] = None):
        ...

    def assert_requested_with(self, expected: Optional[ExpectedHTTPRequest] = None, **kwargs):
        """
        Asserts that the latest request recorded matches an expected request
        Args:
            expected: an expected request.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedHTTPRequest.
        """
        if expected and kwargs:
            raise TypeError('method can be called with either expected or keyword args, but not both')
        if not expected:
            if not kwargs:
                raise TypeError('either expected or keyword args must be provided')
            expected = ExpectedHTTPRequest(**kwargs)

        if not self:
            raise AssertionError('No requests were made')
        match = expected.matches(self[-1])
        if not match:
            raise AssertionError(str(match))

    @overload
    def assert_requested_once_with(self, expected: ExpectedHTTPRequest):
        ...

    @overload
    def assert_requested_once_with(self, *, headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                                   headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                                   path: Optional[Union[str, Pattern[str]]] = None,
                                   path_params: Optional[Mapping[str, Any]] = None,
                                   path_params_submap: Optional[Mapping[str, Any]] = None,
                                   query_params: Optional[Mapping[str, Collection[str]]] = None,
                                   query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                                   method: Optional[str] = None, body: Optional[bytes] = None,
                                   text: Optional[str] = None, json: Any = _missing,
                                   content_predicate:
                                   Optional[Union[Callable[[bytes], bool], Tuple[Callable[[bytes], Any], Any]]] = None):
        ...

    def assert_requested_once_with(self, expected: Optional[ExpectedHTTPRequest] = None, **kwargs):
        """
        Asserts that there is only one request, and that it matches an expected request
        Args:
            expected: an expected request.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedHTTPRequest.
        """
        if expected and kwargs:
            raise TypeError('method can be called with either expected or keyword args, but not both')
        if not expected:
            if not kwargs:
                raise TypeError('either expected or keyword args must be provided')
            expected = ExpectedHTTPRequest(**kwargs)

        if not self:
            raise AssertionError('No requests were made')
        if len(self) > 1:
            raise AssertionError('Multiple requests were made:'
                                 + ''.join(f'\n\t{existing}' for existing in self))
        match = expected.matches(self[0])
        if not match:
            raise AssertionError(str(match))

    @overload
    def assert_any_request(self, expected: ExpectedHTTPRequest):
        ...

    @overload
    def assert_any_request(self, *, headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                           headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                           path: Optional[Union[str, Pattern[str]]] = None,
                           path_params: Optional[Mapping[str, Any]] = None,
                           path_params_submap: Optional[Mapping[str, Any]] = None,
                           query_params: Optional[Mapping[str, Collection[str]]] = None,
                           query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                           method: Optional[str] = None, body: Optional[bytes] = None,
                           text: Optional[str] = None, json: Any = _missing,
                           content_predicate:
                           Optional[Union[Callable[[bytes], bool], Tuple[Callable[[bytes], Any], Any]]] = None):
        ...

    def assert_any_request(self, expected: Optional[ExpectedHTTPRequest] = None, **kwargs):
        """
        Asserts that at least one request recorded matches an expected request
        Args:
            expected: an expected request.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedHTTPRequest.
        """
        if expected and kwargs:
            raise TypeError('method can be called with either expected or keyword args, but not both')
        if not expected:
            if not kwargs:
                raise TypeError('either expected or keyword args must be provided')
            expected = ExpectedHTTPRequest(**kwargs)

        if not self:
            raise AssertionError('No requests were made')
        whynots: List[Tuple[RecordedHTTPRequest, MismatchReason]] = []
        for req in self:
            match = expected.matches(req)
            if match:
                return
            assert isinstance(match, MismatchReason)
            whynots.append((req, match))
        raise AssertionError(f'expected request {expected}, but no requests match:',
                             ''.join(f'\n\t {existing}- {whynot}' for (existing, whynot) in whynots))

    def assert_has_requests_any_order(self, *expected_requests: ExpectedHTTPRequest):
        """
        Asserts that all of the expected requests exclusively match a one of the recorded requests, in any order.
        Args:
            *expected_requests: the expected requests.
        """
        if not expected_requests:
            raise TypeError('at least one expected request must be provided')
        if len(self) < len(expected_requests):
            raise AssertionError(f"expected {len(expected_requests)} sequential requests, but found {len(self)}: "
                                 f"{self}")
        # this problem is equivalent to a maximum bipartite matching in a graph, where one set holds the recorded
        # requests, and the other, the expected requests. We construct an igraph object to represent the problem.
        edges = []
        # since recorded requests are unhashable, we store the why_nots for requests by their ids
        why_nots: Dict[ExpectedHTTPRequest, Dict[int, MismatchReason]] = {expected: {} for expected in
                                                                          expected_requests}
        for i, expected in enumerate(expected_requests):
            for j, request in enumerate(self):
                match = expected.matches(request)
                if match:
                    edges.append((i, j + len(expected_requests)))
                else:
                    assert isinstance(match, MismatchReason)
                    why_nots[expected][id(request)] = match
        graph = Graph.Bipartite([0] * len(expected_requests) + [1] * len(self), edges)

        max_matching = graph.maximum_bipartite_matching()
        unmatched_strs = []
        for i, expected in enumerate(expected_requests):
            if not max_matching.is_matched(i):
                unmatched_strs.append(f'could not match expected request {expected}:'
                                      + ''.join((f'\n\t{req}- {why_nots[expected][id(req)]}'
                                                 if id(req) in why_nots[expected]
                                                 else f'\n\t{req}- matched expected request '
                                                      + str(expected_requests[max_matching.match_of(node_index)]))
                                                for node_index, req in enumerate(self, len(expected_requests))))
        if unmatched_strs:
            raise AssertionError('\n'.join(unmatched_strs))

    def assert_has_requests(self, *expected_requests: ExpectedHTTPRequest):
        """
        Asserts that all of the expected requests exclusively match a one of the recorded requests, in sequential order.
        Args:
            *expected_requests: the expected requests.
        Notes:
            The matched requests must be sequential relative to the expected requests, but they needn't be contiguous.
             This means that if requests A,B are expected, then the recorded request sequence A,C,B matches it.
        """
        if not expected_requests:
            raise TypeError('at least one expected request must be provided')

        if len(self) < len(expected_requests):
            raise AssertionError(f"could not find request to match {expected_requests[0]}")
        # match greedily
        expected_iter = iter(expected_requests)
        next_expected = next(expected_iter)
        for request in self:
            match = next_expected.matches(request)
            if match:
                try:
                    next_expected = next(expected_iter)
                except StopIteration:
                    return True
        raise AssertionError(f"could not find request to match {next_expected}")



import json
import os
import sys
from contextlib import contextmanager
from json import JSONDecodeError
from typing import Optional, TextIO

from docker import DockerClient
from docker.errors import BuildError, DockerException, ImageNotFound

from yellowbox.utils import _get_spinner


class DockerfileParseException(BuildError):
    pass


@contextmanager
def build_image(docker_client: DockerClient, image_name: str, remove_image: bool = True,
                file: Optional[TextIO] = sys.stderr, spinner: bool = True, **kwargs):
    """
    Create a docker image (similar to docker build command)
    At the end, deletes the image (using rmi command)
    Args:
        docker_client: DockerClient to be used to create the image
        image_name: Name of the image to be created
        remove_image: boolean, whether or not to delete the image at the end, default as True
        file: a file-like object (stream); defaults to the current sys.stderr. if set to None, will disable printing
        spinner: boolean, whether or not to use spinner (default as True), note that this param is set to False in
        case `file` param is not None
    """
    if file is None:
        file = open(os.devnull, 'w')
    else:
        spinner = False  # spinner splits into multiple lines in case stream is being printed at the same time
    image_tag = f'{image_name}:test'
    yaspin_spinner = _get_spinner(spinner)
    with yaspin_spinner(f'Creating image {image_tag}...'):
        kwargs = {'tag': image_tag, 'rm': True, 'forcerm': True, **kwargs}
        build_log = docker_client.api.build(**kwargs)
        for msg_b in build_log:
            msgs = str(msg_b, 'utf-8').splitlines()
            for msg in msgs:
                try:
                    parse_msg = json.loads(msg)
                except JSONDecodeError:
                    raise DockerException('error at build logs')
                s = parse_msg.get('stream')
                if s:
                    print(s, end='', flush=True, file=file)
                else:
                    # runtime errors
                    error_detail = parse_msg.get('errorDetail')
                    # parse errors
                    error_msg = parse_msg.get('message')
                    # steps of the image creation
                    status = parse_msg.get('status')
                    # end of process, will contain the ID of the temporary container created at the end
                    aux = parse_msg.get('aux')
                    if error_detail is not None:
                        raise BuildError(reason=error_detail, build_log=None)
                    elif error_msg is not None:
                        raise DockerfileParseException(reason=error_msg, build_log=None)
                    elif status is not None:
                        print(status, end='', flush=True, file=file)
                    elif aux is not None:
                        print(aux, end='', flush=True, file=file)
                    else:
                        raise DockerException(parse_msg)
        yield image_tag
        if remove_image:
            try:
                docker_client.api.remove_image(image_tag)
            except ImageNotFound:
                # if the image was already deleted
                pass



from __future__ import annotations

from dataclasses import dataclass
from itertools import count
from time import perf_counter, sleep
from typing import Callable, Iterable, Optional, Tuple, Type, TypeVar, Union

_T = TypeVar('_T')


@dataclass
class RetrySpec:
    """
    Specifications for a repeated attempts af an arbitrary action that might fail.
    """

    interval: float = 2
    """
    Time between attempts in seconds.
    """
    attempts: Optional[int] = None
    """
    Max number of attempts. If None, infinite attempts are made.
    """
    timeout: Optional[float] = None
    """
    A timeout for all the attempts (including the interval) combined.
    """

    def __post_init__(self):
        if self.attempts is self.timeout is None:
            raise ValueError('RetrySpec must have either a timeout or attempts')

    def retry(self, func: Callable[[], _T],
              exceptions: Union[Type[Exception], Tuple[Type[Exception], ...]]) -> _T:
        """
        Retry running func until it succeeds

        Args:
            func: Function to run without arguments.
            exceptions: Single or iterable of exception types to catch. Exceptions raised by `func` not residing
            inside this iterable will be propagated.

        Returns:
            The return value of `func`'s first successful call.

        Raises:
            Any exception raised by `func` if max attempts or timeout were reached or  if the exception
            wasn't specified in exceptions list.

        """
        attempt_iterator: Iterable
        if self.attempts is None:
            attempt_iterator = count()
        elif self.attempts < 1:
            raise ValueError("Attempts must be greater than zero.")
        else:
            attempt_iterator = range(self.attempts - 1)

        if self.timeout:
            time_limit = perf_counter() + self.timeout
        else:
            time_limit = float('inf')

        for _ in attempt_iterator:
            try:
                return func()
            except exceptions:
                if perf_counter() > time_limit:
                    raise
            sleep(self.interval)
        return func()



from contextlib import closing
from typing import ContextManager, Optional, Tuple, Union, cast
from uuid import uuid1

from docker import DockerClient
from kafka import KafkaConsumer, KafkaProducer
from kafka.errors import KafkaError

from yellowbox.containers import SafeContainerCreator, get_ports
from yellowbox.networks import anonymous_network
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import RunMixin, SingleEndpointService
from yellowbox.utils import get_free_port

__all__ = ['KafkaService']


class KafkaService(SingleEndpointService, RunMixin):
    def __init__(self, docker_client: DockerClient, tag_or_images: Union[str, Tuple[str, str]] = 'latest',
                 inner_port=0, outer_port=0, **kwargs):
        self.inner_port = inner_port or get_free_port()
        self.outer_port = outer_port or get_free_port()
        if isinstance(tag_or_images, str):
            zookeeper_image = f"bitnami/zookeeper:{tag_or_images}"
            broker_image = f"bitnami/kafka:{tag_or_images}"
        else:
            zookeeper_image, broker_image = tag_or_images

        # broker must have a known alias at creation time
        self.static_broker_alias = f'broker-{uuid1()}'

        creator = SafeContainerCreator(docker_client)

        self.zookeeper = creator.create_and_pull(
            zookeeper_image, detach=True,
            publish_all_ports=True,
            environment={
                'ZOOKEEPER_CLIENT_PORT': '2181',
                'ZOOKEEPER_TICK_TIME': '2000',
                'ALLOW_ANONYMOUS_LOGIN': 'yes',
            })

        self.broker = creator.create_and_pull(
            broker_image,
            ports={
                str(self.outer_port): ('0.0.0.0', self.outer_port),
                str(self.inner_port): ('0.0.0.0', self.inner_port)
            },
            publish_all_ports=True,
            detach=True,
            environment={
                'KAFKA_CFG_ADVERTISED_LISTENERS': f'INNER://{self.static_broker_alias}:{self.inner_port},'
                                                  f'OUTER://localhost:{self.outer_port}',
                "KAFKA_CFG_ZOOKEEPER_CONNECT": "zk/2181",
                'ALLOW_PLAINTEXT_LISTENER': 'yes',
                'KAFKA_CFG_LISTENER_SECURITY_PROTOCOL_MAP': 'INNER:PLAINTEXT,OUTER:PLAINTEXT',
                'KAFKA_INTER_BROKER_LISTENER_NAME': 'INNER',
                'KAFKA_CFG_LISTENERS': f'INNER://:{self.inner_port},OUTER://:{self.outer_port}'
            })

        self.network = anonymous_network(docker_client)
        self.network.connect(self.zookeeper, aliases=["zk"])
        self.network.connect(self.broker, aliases=[self.static_broker_alias])
        super().__init__((self.zookeeper, self.broker), **kwargs)

    def connection_port(self):
        self.broker.reload()
        ports = get_ports(self.broker)
        return ports[self.outer_port]

    def consumer(self, **kwargs) -> ContextManager[KafkaConsumer]:
        port = self.connection_port()
        return cast(
            'ContextManager[KafkaConsumer]',
            closing(KafkaConsumer(bootstrap_servers=[f'localhost:{port}'], security_protocol="PLAINTEXT", **kwargs))
        )

    def producer(self, **kwargs) -> ContextManager[KafkaProducer]:
        port = self.connection_port()
        return cast(
            'ContextManager[KafkaProducer]',
            closing(KafkaProducer(bootstrap_servers=[f'localhost:{port}'], security_protocol="PLAINTEXT", **kwargs))
        )

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=20)
        with retry_spec.retry(self.consumer, (KafkaError, ConnectionError, ValueError)):
            pass
        return self

    def stop(self, signal='SIGKILL'):
        # difference in default signal
        self.network.disconnect(self.broker)
        self.network.disconnect(self.zookeeper)
        self.network.remove()
        super().stop(signal)

    def connect(self, network, *, aliases=(), **kwargs):
        if not isinstance(aliases, list):
            aliases = list(aliases)
        aliases.append(self.static_broker_alias)
        kwargs['aliases'] = aliases
        return super().connect(network, **kwargs)

    @property
    def _single_endpoint(self):
        return self.broker



from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Lock, Thread
from types import new_class
from typing import Callable, ClassVar, DefaultDict, List, Mapping, NamedTuple, Optional, Pattern, Set, Type, Union, cast
from urllib.parse import ParseResult, parse_qs, urlparse

import requests
from requests import ConnectionError, HTTPError

from yellowbox.retry import RetrySpec
from yellowbox.service import YellowService
from yellowbox.utils import docker_host_name

__all__ = ['HttpService', 'RouterHTTPRequestHandler']
SideEffectResponse = Union[bytes, str, int, 'RouterHTTPRequestHandler']
SideEffect = Union[Callable[['RouterHTTPRequestHandler'], None],
                   Callable[['RouterHTTPRequestHandler'], SideEffectResponse],
                   SideEffectResponse]


class RoutedHandler(NamedTuple):
    method: str
    name: str
    route: Union[Pattern[str], str]
    callback: Callable[[RouterHTTPRequestHandler], None]

    def route_match(self, path: str):
        if isinstance(self.route, str):
            return self.route == path
        return self.route.fullmatch(path)


class RouterHTTPRequestHandler(BaseHTTPRequestHandler):
    """
    A BaseHTTPRequestHandler that allows adding and deleting routed handlers.
    Also contains some utility argument parsing.
    """
    _parse_url: ParseResult
    _body: Optional[bytes]

    routes_by_method: ClassVar[DefaultDict[str, Set[RoutedHandler]]]
    route_lock: ClassVar[Lock]

    def body(self) -> Optional[bytes]:
        try:
            return self._body
        except AttributeError:
            raw_body_len = self.headers['Content-Length']
            if raw_body_len is None:
                self._body = None
            else:
                length = int(raw_body_len)
                self._body = self.rfile.read(length)
            return self._body

    def path_params(self, **kwargs) -> Mapping[str, List[str]]:
        """
        Extract the path parameters from the query
        Args:
            **kwargs: forwarded to urllib.parse.parse_qs

        Returns:
            A mapping from parameter name to a list of values provided

        """
        parsed = self.parse_url()
        return parse_qs(parsed.query, **kwargs)

    def parse_url(self) -> ParseResult:
        try:
            return self._parse_url
        except AttributeError:
            self._parse_url = urlparse(self.path)
            return self._parse_url

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.routes_by_method = defaultdict(set)
        cls.route_lock = Lock()

    @classmethod
    def add_route(cls, handler: RoutedHandler):
        with cls.route_lock:
            cls.routes_by_method[handler.method].add(handler)

    @classmethod
    def del_route(cls, handler: RoutedHandler):
        with cls.route_lock:
            cls.routes_by_method[handler.method].remove(handler)

    def _do(self):
        """
        A generic handler for all http methods.
        Filters through all added routes and calls the appropriate callback, if one is found.
        """
        parsed = self.parse_url()

        candidates = tuple(self.routes_by_method.get(self.command, ()))
        matched_candidates = []
        for candidate in candidates:
            match = candidate.route_match(parsed.path)
            if match:
                matched_candidates.append((candidate, match))
        if not matched_candidates:
            self.send_error(404, 'mock server matched no routes')
            self.end_headers()
        elif len(matched_candidates) > 1:
            self.send_error(500, 'mock server matched multiple routes: '
                            + ', '.join(c.name for c, _ in matched_candidates))
            self.end_headers()
        else:
            (routed, match), = matched_candidates
            self.log_message(f'routed to {routed.name}')
            self.match = match
            routed.callback(self)

    def __getattr__(self, item: str):
        if item.startswith('do_'):
            return self._do
        raise AttributeError(item)


class HttpService(YellowService):
    """
    The HttpService class is used to mock http servers. Although it is a YellowService,
    it does not wrap a docker container, rather, it wraps a standard library HTTPServer, with a server
    thread running in the background. The server uses an internal router that maps paths to callbacks.

    Example usage:
    >>> with HttpService().start() as service:
    ...   @service.patch_route('GET', '/hello/world')
    ...   def hello_world(handler: RouterHTTPRequestHandler):
    ...      return "hi there"
    ...   # hello_world is now a context manager
    ...   with hello_world:
    ...      # within this scope, the path "/hello/world" will return a 200 response with the body "hi there"
    ...      assert requests.get(service.local_url+"/hello/world").text == "hi there"
    ...   # routes can also be set without a function
    ...   with service.patch_route('GET', '/meaning_of_life', '42'):
    ...      assert requests.get(service.local_url+"/meaning_of_life").content == b'42'
    """

    def __init__(self, host='0.0.0.0', port=0, name='anonymous_yellowbox_HTTPService'):
        self.router_cls = cast(Type[RouterHTTPRequestHandler],
                               new_class(name + '_RequestHandler', (RouterHTTPRequestHandler,)))
        self.server = HTTPServer((host, port), self.router_cls)
        self.server_thread = Thread(name=name + '_thread', target=self.server.serve_forever,
                                    daemon=True)

    @property
    def server_port(self):
        return self.server.server_port

    @property
    def local_url(self):
        return f'http://127.0.0.1:{self.server_port}'

    @property
    def container_url(self):
        return f'http://{docker_host_name}:{self.server_port}'

    @staticmethod
    def _to_callback(side_effect: SideEffect):
        def _respond(handler: RouterHTTPRequestHandler, response: SideEffectResponse):
            if response is handler:
                return  # Assuming the user already handled the response
            if isinstance(response, int):
                handler.send_error(response)
                handler.end_headers()
                return
            handler.send_response(200)
            handler.end_headers()
            if isinstance(response, str):
                response = bytes(response, 'ascii')
            if isinstance(response, bytes):
                handler.wfile.write(response)
            else:
                raise TypeError(f"got response of type {type(response)}, type must be RouterHTTPRequestHandler, "
                                f"int, str or bytes")

        def callback(handler: RouterHTTPRequestHandler):
            if callable(side_effect):
                result = side_effect(handler)
                _respond(handler, result)
            else:
                _respond(handler, cast(SideEffectResponse, side_effect))

        return callback

    def patch_route(self, method, route: Union[str, Pattern[str]],
                    side_effect: SideEffect = ...,  # type: ignore[assignment]
                    name: Optional[str] = None):
        """
        Create a context manager that temporarily adds a route handler to the service.

        Args:
            method: The request method to add the route to.
            route: The route to attach the side effect to, all routes must begin with a slash "/".
                Alternatively, The route may be a regex pattern, in which case the request path must fully match it,
                the match object is then stored in RouterHTTPRequestHandler.match, to be used by a side-effect callable.
            side_effect: The callback or result to return for the route. Accepts any of the following types:
                * int: to return the value as the HTTP status code, without a body.
                * bytes: to return 200, with the value as the response body.
                * str: invalid if the value is non-ascii, return 200 with the value, translated to bytes, as
                 the response body.
                * callable: Must accept a RouterHTTPRequestHandler. May return any of the above types, or
                  RouterHTTPRequestHandler to handle the response directly with the RouterHTTPRequestHandler.
            name:
                An optional name for the routed handler, to be used while logging. If missing, a suitable name is
                extracted from the side effect or route.

        Returns:
            A context manager that will enable the route upon entry and disable it upon exit.

        Notes:
            This method can be used as a decorator by not specifying `side_effect`
        """
        if side_effect is ...:
            return partial(self.patch_route, method, route, name=name)

        @contextmanager
        def _helper():
            nonlocal name
            callback = self._to_callback(side_effect)
            name = name or getattr(side_effect, '__name__', None) or str(route)
            handler = RoutedHandler(method, name, route, callback)
            self.router_cls.add_route(handler)
            try:
                yield handler
            finally:
                self.router_cls.del_route(handler)

        return _helper()

    def start(self, retry_spec: Optional[RetrySpec] = None):
        with self.patch_route('GET', '/health', 200):
            self.server_thread.start()
            retry_spec = retry_spec or RetrySpec(attempts=10)
            retry_spec.retry(
                lambda: requests.get(self.local_url + '/health').raise_for_status(),
                (ConnectionError, HTTPError)
            )
        return super(HttpService, self).start()

    def stop(self):
        self.server.shutdown()
        self.server_thread.join()

    def is_alive(self):
        return self.server_thread.is_alive()



from functools import update_wrapper
from typing import TYPE_CHECKING, Callable, ClassVar, Generic, Optional, TypeVar, overload

from yellowbox.extras.webserver.endpoints import (
    HTTP_SIDE_EFFECT, METHODS, WS_SIDE_EFFECT, MockHTTPEndpoint, MockWSEndpoint, http_endpoint, ws_endpoint
)

T = TypeVar('T')


class EndpointTemplate(Generic[T]):
    """
    A template for a generic endpoint. To be created before any instance of a subclass of webserver, but to be added to
     all instances.
    """
    constructor: ClassVar[Callable[..., T]]

    def __init__(self, *args, side_effect_method, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.side_effect_method = side_effect_method
        update_wrapper(self, side_effect_method)

    def construct(self, instance) -> T:
        if hasattr(self.side_effect_method, '__get__'):
            side_effect = self.side_effect_method.__get__(instance, type(instance))
        else:
            side_effect = self.side_effect_method
        return self.constructor(*self.args, side_effect, **self.kwargs)

    if TYPE_CHECKING:
        def __get__(self, instance, owner) -> T:
            ...


class HTTPEndpointTemplate(EndpointTemplate[MockHTTPEndpoint]):
    """
    A subclass of EndpointTemplate that is used to create an HTTP endpoint.
    """
    constructor = staticmethod(http_endpoint)  # type: ignore[assignment]


class WSEndpointTemplate(EndpointTemplate[MockWSEndpoint]):
    """
    A subclass of EndpointTemplate that is used to create a websocket endpoint.
    """
    constructor = staticmethod(ws_endpoint)  # type: ignore[assignment]


@overload
def class_http_endpoint(methods: METHODS, rule_string: str, *, auto_read_body: bool = True,
                        forbid_head_verb: bool = True) -> Callable[[HTTP_SIDE_EFFECT], HTTPEndpointTemplate]: ...


@overload
def class_http_endpoint(methods: METHODS, rule_string: str, side_effect: HTTP_SIDE_EFFECT, *,
                        auto_read_body: bool = True, forbid_implicit_head_verb: bool = True) -> HTTPEndpointTemplate:
    ...


def class_http_endpoint(methods: METHODS, rule_string: str, side_effect: Optional[HTTP_SIDE_EFFECT] = None, **kwargs):
    """
    Creates an HTTP endpoint template. Declare this as a class variable in your webserver subclass to automatically add
     the endpoint to all instances. Can be used as a decorator.
    Args:
        methods: forwarded to MockHTTPEndpoint
        rule_string: forwarded to MockHTTPEndpoint
        side_effect: forwarded to MockHTTPEndpoint
        **kwargs: forwarded to MockHTTPEndpoint

    Returns:
        A new http endpoint template

    """
    def ret(side_effect_method):
        return HTTPEndpointTemplate(methods, rule_string, side_effect_method=side_effect_method, **kwargs)

    if side_effect is not None:
        return ret(side_effect)
    return ret


@overload
def class_ws_endpoint(rule_string: str) -> Callable[[WS_SIDE_EFFECT], WSEndpointTemplate]: pass


@overload
def class_ws_endpoint(rule_string: str, side_effect: WS_SIDE_EFFECT) -> WSEndpointTemplate: pass


def class_ws_endpoint(rule_string: str, side_effect: Optional[WS_SIDE_EFFECT] = None):
    """
    Creates a websocket endpoint template. Declare this as a class variable in your webserver subclass to automatically
     add the endpoint to all instances. Can be used as a decorator.
    Args:
        rule_string: forwarded to MockWSEndpoint
        side_effect: forwarded to MockWSEndpoint

    Returns:
        A new websocket endpoint template

    """
    def ret(side_effect_method):
        return WSEndpointTemplate(rule_string, side_effect_method=side_effect_method)

    if side_effect is not None:
        return ret(side_effect)
    return ret



from __future__ import annotations

from contextlib import contextmanager
from logging import Filter, getLogger
from typing import Awaitable, Callable, Iterable, TypeVar, Union

from starlette.responses import Response


def reason_is_ne(field: str, expected, got) -> str:
    """
    Create a string that is describes two values being unequal
    Args:
        field: the name of the mismatched field
        expected: the expected value
        got: the actual value
    """
    return f'{field} mismatch: expected {expected}, got {got}'


class MismatchReason(str):
    """
    A falsish object, signifying a failure to match, with a reason.
    """

    def __bool__(self):
        return False


class MuteFilter(Filter):
    """
    A simple filter that silences all logs
    """

    def filter(self, record) -> bool:
        return False


@contextmanager
def mute_uvicorn_log():
    """
    A context manager to silence all log messages from the uvicorn.error logger while in its scope
    """
    logger = getLogger('uvicorn.error')
    filter_ = MuteFilter()
    logger.addFilter(filter_)
    yield
    logger.removeFilter(filter_)


T = TypeVar('T')


def iter_side_effects(side_effects: Iterable[Union[Callable[..., Awaitable[T]], T]]) -> Callable[..., Awaitable[T]]:
    """
    Args:
        side_effects: An iterable of side effects.

    Returns:
        A side effect that varies from call to call. On the first call, delegating to the first side effect on
         side_effects, on the second to the second, and so on.

    Notes:
        This function respects starlette responses in that if it encounters one, it will simply return it instead of
         delegating to it.
    """
    tor = iter(side_effects)

    async def ret(*args, **kwargs):
        next_side_effect = next(tor)
        if isinstance(next_side_effect, Response):  # responses are async callable smh
            return next_side_effect
        return await next_side_effect(*args, **kwargs)

    return ret



from typing import Optional

from docker import DockerClient
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

from yellowbox import RetrySpec, RunMixin
from yellowbox.containers import create_and_pull, get_ports
from yellowbox.subclasses import SingleContainerService

__all__ = ['PostgreSQLService', 'POSTGRES_INTERNAL_PORT']

from yellowbox.utils import docker_host_name

POSTGRES_INTERNAL_PORT = 5432


class PostgreSQLService(SingleContainerService, RunMixin):
    """
    A postgresSQL service
    """

    def __init__(self, docker_client: DockerClient, image='postgres:latest', *, user='postgres',
                 password='guest', default_db: str = None, **kwargs):
        """
        Args:
            user: the Name of the default user for the database
            password: The password of the default user for the database
            default_db: The name of the default database. Defaults to the user name.
        """
        if default_db is None:
            default_db = user

        self.user = user
        self.password = password
        self.default_db = default_db
        super().__init__(create_and_pull(
            docker_client, image, publish_all_ports=True, detach=True, environment={
                'POSTGRES_USER': user,
                'POSTGRES_PASSWORD': password,
                'POSTGRES_DB': default_db
            }
        ), **kwargs)

    def external_port(self):
        return get_ports(self.container)[POSTGRES_INTERNAL_PORT]

    def local_connection_string(self, dialect: str = 'postgresql', driver: str = None, database: str = None):
        """
        Generate an sqlalchemy-style connection string to the database in the service from the docker host.
        Args:
            dialect: The dialect of the sql server.
            driver: additional driver for sqlalchemy to use.
            database: the name of the database to connect to. Defaults to the service's default database.
        """
        database = database or self.default_db

        if driver is not None:
            dialect += '+' + driver

        return f'{dialect}://{self.user}:{self.password}@localhost:{self.external_port()}/{database}'

    def container_connection_string(self, hostname: str, dialect: str = 'postgresql', driver: str = None,
                                    database: str = None):
        """
        Generate an sqlalchemy-style connection string to the database in the service from another container on a
         common network.
        Args:
            hostname: the alias of the container.
            dialect: The dialect of the sql server.
            driver: additional driver for sqlalchemy to use.
            database: the name of the database to connect to. Defaults to the service's default database.
        """
        database = database or self.default_db

        if driver is not None:
            dialect += '+' + driver

        return f'{dialect}://{self.user}:{self.password}@{hostname}:{POSTGRES_INTERNAL_PORT}/{database}'

    def host_connection_string(self, dialect: str = 'postgresql', driver: str = None, database: str = None):
        """
        Generate an sqlalchemy-style connection string to the database in the service from another container.
        Args:
            dialect: The dialect of the sql server.
            driver: additional driver for sqlalchemy to use.
            database: the name of the database to connect to. Defaults to the service's default database.
        """
        database = database or self.default_db

        if driver is not None:
            dialect += '+' + driver

        return f'{dialect}://{self.user}:{self.password}@{docker_host_name}:{self.external_port()}/{database}'

    def engine(self, **kwargs):
        """
        Create an sqlalchemy Engine connected to the service's default db.
        """
        cs = self.local_connection_string()
        return create_engine(cs, **kwargs)

    def connection(self, **kwargs):
        """
        Create an sqlalchemy Connection connected to the service's default db.
        """
        return self.engine().connect(**kwargs)

    def start(self, retry_spec: Optional[RetrySpec] = None):
        retry_spec = retry_spec or RetrySpec(attempts=20)
        super().start(retry_spec)

        def connect():
            with self.connection():
                return

        retry_spec.retry(connect, OperationalError)
        return self

    def stop(self, signal='SIGINT'):
        # change in default
        return super().stop(signal)



from contextlib import contextmanager
from typing import ContextManager, Optional
from urllib.parse import quote

import requests
from docker import DockerClient
from pika import BlockingConnection, ConnectionParameters, PlainCredentials
from pika.exceptions import AMQPConnectionError

from yellowbox.containers import create_and_pull, get_ports, upload_file
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import RunMixin, SingleContainerService

__all__ = ['RabbitMQService', 'RABBIT_DEFAULT_PORT', 'RABBIT_HTTP_API_PORT']

RABBIT_DEFAULT_PORT = 5672
RABBIT_HTTP_API_PORT = 15672


class RabbitMQService(SingleContainerService, RunMixin):
    def __init__(self, docker_client: DockerClient, image='rabbitmq:latest', *, user="guest", password="guest",
                 virtual_host="/", **kwargs):
        self.user = user
        self.password = password
        self.virtual_host = virtual_host
        super().__init__(create_and_pull(
            docker_client, image, publish_all_ports=True, detach=True,
            ports={RABBIT_HTTP_API_PORT: 0},  # Forward management port by default.
        ), **kwargs)

        upload_file(self.container, '/etc/rabbitmq/rabbitmq.conf',
                    f'''
                    default_pass = {password}
                    default_user = {user}
                    default_vhost = {virtual_host}
                    loopback_users = none
                    '''.encode())

    def connection_port(self):
        return get_ports(self.container)[RABBIT_DEFAULT_PORT]

    def connection(self, **kwargs):
        credentials = PlainCredentials(self.user, self.password)
        connection_params = ConnectionParameters(
            'localhost', self.connection_port(),
            credentials=credentials, virtual_host=self.virtual_host,
            **kwargs
        )
        return BlockingConnection(connection_params)

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=30)
        conn = retry_spec.retry(self.connection, (AMQPConnectionError, ConnectionError))
        conn.close()
        return self

    def management_url(self):
        try:
            return f"http://localhost:{get_ports(self.container)[RABBIT_HTTP_API_PORT]}/"
        except KeyError as exc:
            raise RuntimeError("Management is not enabled.") from exc

    def enable_management(self):
        if not self.is_alive():
            raise RuntimeError("Must be used on an already-running container.")

        if RABBIT_HTTP_API_PORT not in get_ports(self.container):
            raise RuntimeError("Container must have the management port exposed.")

        self.container.exec_run("rabbitmq-plugins enable rabbitmq_management")

    def reset_state(self, force_queue_deletion=False):
        try:
            management_url = self.management_url()
        except RuntimeError as e:
            raise RuntimeError('management must be enabled for clean_slate') from e

        queues_url = management_url + 'api/queues'
        replies = requests.get(queues_url, auth=(self.user, self.password))
        replies.raise_for_status()
        extant_queues = replies.json()
        delete_params = {}
        if not force_queue_deletion:
            delete_params['if-unused'] = 'true'
        for queue in extant_queues:
            name = quote(queue['name'], safe='')
            vhost = quote(queue['vhost'], safe='')
            requests.delete(
                management_url + f'api/queues/{vhost}/{name}',
                auth=(self.user, self.password), params=delete_params
            ).raise_for_status()

    @classmethod
    @contextmanager
    def run(cls, docker_client: DockerClient, *, enable_management=False, **kwargs):
        cmg: ContextManager[RabbitMQService] = super().run(docker_client, **kwargs)
        with cmg as ret:
            if enable_management:
                ret.enable_management()
            yield ret

    def stop(self, signal='SIGKILL'):
        # change in default
        return super().stop(signal)



import platform
from contextlib import AbstractContextManager, closing, contextmanager, nullcontext
from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOL_SOCKET, socket
from typing import Callable, TypeVar

from yaspin import yaspin

_T = TypeVar('_T')
_SPINNER_FAILMSG = "💥 "
_SPINNER_SUCCESSMSG = "✅ "


@contextmanager
def _spinner(text):
    with yaspin(text=text) as spinner:
        try:
            yield
        except Exception:
            spinner.fail(_SPINNER_FAILMSG)
            raise
        spinner.ok(_SPINNER_SUCCESSMSG)


def _get_spinner(real=True) -> Callable[[str], AbstractContextManager]:
    if not real:
        return lambda text: nullcontext()
    return _spinner


def get_free_port():
    with closing(socket(AF_INET, SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        return s.getsockname()[1]


if platform.system() == "Linux" and ('microsoft' not in platform.uname().release.lower()):  # catch WSL
    docker_host_name = '172.17.0.1'
else:
    docker_host_name = 'host.docker.internal'



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Collection, Dict, Iterable, List, Mapping, Optional, Pattern, Sequence, Tuple, Union, overload

from starlette.requests import HTTPConnection
from starlette.websockets import WebSocket

from yellowbox.extras.webserver.request_capture import ScopeExpectation
from yellowbox.extras.webserver.util import MismatchReason, reason_is_ne


class WebSocketRecorder(WebSocket):
    """
    A subclass of starlette.Websocket that records the conversation in a transcript
    messages are parsed as per https://asgi.readthedocs.io/en/latest/specs/www.html#websocket
    """

    def __init__(self, scope, receive, send, sinks: Sequence[RecordedWSTranscripts]):
        """
        Args:
            scope: forwarded to Websocket
            receive: forwarded to Websocket
            send: forwarded to Websocket
            sinks: a list of transcripts to add the transcript to
        """
        super().__init__(scope, receive, send)
        self.transcript: RecordedWSTranscript = RecordedWSTranscript.from_connection(self)
        for sink in sinks:
            sink.append(self.transcript)

    async def receive(self):
        message = await super().receive()
        message_type = message.get('type')
        if message_type == "websocket.receive":
            # by asgi uvicorn implementation (def asgi_receive), a message will always contain exactly one of
            #  'text' or 'data'
            data = message.get('text')
            if data is None:
                data = message.get('bytes')
            self.transcript.append(RecordedWSMessage(data, Sender.Client))
        elif message_type == "websocket.disconnect":
            self.transcript.close = (Sender.Client, message.get('code', 1000))
        return message

    async def send(self, message):
        message_type = message["type"]
        if message_type == "websocket.accept":
            self.transcript.accepted = True
        elif message_type == "websocket.close":
            self.transcript.close = (Sender.Server, message.get('code', 1000))
        elif message_type == "websocket.send":
            # by asgi uvicorn implementation (def asgi_receive), a message will always contain exactly one of
            #  'text' or 'data'
            data = message.get('text')
            if data is None:
                data = message.get('bytes')
            self.transcript.append(RecordedWSMessage(data, Sender.Server))
        await super().send(message)


async def recorder_websocket_endpoint(scope, receive, send, *, function, sinks: Sequence[RecordedWSTranscripts]):
    """
    An entrypoint app that records websocket conversations
    Args:
        scope: the http scope of the conversation
        receive: forwarded to the Websocket
        send: forwarded to the Websocket
        function: the function to call the endpoint with
        sinks: a list of transcripts that the record should be added to
    """
    # a partial version of this function will be the app endpoint of all our mock websocket routes
    ws = WebSocketRecorder(scope, receive, send, sinks=sinks)
    return await function(ws)


class Sender(Enum):
    """
    A sender that can send messages in a transcript. A sender can be called to create an expected message.

    Examples:
        >>> import re
        ... expected_message = Sender.Server(re.compile('[a-z][a-z][0-9]'))
        ... message = RecordedWSMessage('tk1', Sender.Server)
        ... assert expected_message.matches(message)
    """
    Server = auto()
    Client = auto()

    def __call__(self, data: Union[Pattern[str], Pattern[bytes], str, bytes, ellipsis]) \
            -> ExpectedWSMessage:  # noqa: F821
        """
        Create an expected message, originating from this caller
        Args:
            data: the expected data of the message. Can be either a bytes or string for an exact match, a compiled
             pattern for a full match, or ellipsis to match any data.
        """
        return ExpectedWSMessage(data, self)


@dataclass
class RecordedWSMessage:
    """
    A recorded websocket message, with a sender and value sent
    """
    data: Union[bytes, str]
    sender: Sender

    def __repr__(self):
        return f'RecordedWSMessage({self.data!r}, Sender.{self.sender.name})'


class RecordedWSTranscript(List[RecordedWSMessage]):
    """
    A transcript of a single websocket connection
    """

    def __init__(self, arg: Iterable[RecordedWSMessage], headers: Dict[bytes, List[bytes]], path: str,
                 path_params: Dict[str, Any], query_params: Dict[str, List[str]]):
        """
        Args:
            arg: forwarded to super (List)
            headers: the headers of the connection
            path: the path of the connection request
            path_params: the path params of the connection request, as specified by the route's starlette rule string
            query_params: the query params of the connection request
        """
        super().__init__(arg)
        self.headers = headers
        self.path = path
        self.path_params = path_params
        self.query_params = query_params
        self.accepted: bool = False
        """Whether the message was ever accepted by the server"""
        self.close: Optional[Tuple[Sender, int]] = None
        """The sender who closed the connection, along with the close code"""

    @classmethod
    def from_connection(cls, connection: HTTPConnection):
        """
        Create an empty transcript from a new http connection
        Args:
            connection: the http connection to use
        """
        headers: Dict[bytes, List[bytes]] = {}
        for h_k, h_v in connection.headers.raw:
            h_lst = headers.get(h_k)
            if h_lst is None:
                headers[h_k] = [h_v]
            else:
                h_lst.append(h_v)
        path = connection.url.path
        path_params = connection.path_params
        query_params: Dict[str, List[str]] = {}
        for q_k, q_v in connection.query_params.multi_items():
            q_lst = query_params.get(q_k)
            if q_lst is None:
                query_params[q_k] = [q_v]
            else:
                q_lst.append(q_v)
        return cls((), headers, path, path_params, query_params)


class ExpectedWSMessage:
    """
    A single expected message in a websockets connection
    """

    def __init__(self, data: Union[Pattern[str], Pattern[bytes], str, bytes, ellipsis], sender: Sender):  # noqa: F821
        """
        Args:
            data: the expected data. Can be either a bytes or string for an exact match, a compiled pattern for a full
             match, or ellipsis to match any data.
            sender: The sender to expect the message from.
        """
        self.data = data
        self.sender = sender

    def matches(self, message: RecordedWSMessage) -> Union[bool, MismatchReason]:
        """
        Checks whether a recorded message matches the expectation
        Args:
            message: the recorded message

        Returns:
            True if the message matches, or a mismatch reason otherwise.
        """
        if self.sender != message.sender:
            return MismatchReason(reason_is_ne('sender', self.sender, message.sender))

        if self.data is ...:
            return True
        elif isinstance(self.data, Pattern):
            try:
                match = self.data.fullmatch(message.data)  # type:ignore[arg-type]
            except TypeError:  # this would happen when we try to use byte patterns on strs or vice-versa
                match = None
            if not match:
                return MismatchReason(f'expected data to match pattern {self.data.pattern!r}, got {message.data!r}')
            return True
        else:
            if self.data != message.data:
                return MismatchReason(reason_is_ne('data', self.data, message.data))
        return True

    def __str__(self):
        return f'{self.sender.name}({self.data})'


class ExpectedWSTranscript(ScopeExpectation):
    """
    An expected websocket transcript
    """

    def __init__(self, messages: Sequence[Union[ellipsis, ExpectedWSMessage]] = (...,),  # noqa: F821
                 headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                 headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                 path: Optional[Union[str, Pattern[str]]] = None, path_params: Optional[Mapping[str, Any]] = None,
                 path_params_submap: Optional[Mapping[str, Any]] = None,
                 query_params: Optional[Mapping[str, Collection[str]]] = None,
                 query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                 close: Optional[Tuple[Sender, int]] = None, accepted: Optional[bool] = True):
        """
        Args:
            messages: the expected messages in the transcript, can begin or end ellipsis to signify that any
                number of messages and precede or follow the messages to match.
            headers: If specified, expects the request to have these headers exactly
            headers_submap: If specified expects the request to have at least the headers specified
            path: If specified, expected the request url (after the host) to either be equal to this (in case of string)
             or to fully match the regex pattern provided
            path_params: If specified, expects the request to have exactly these path parameters, as provided by the
             starlette request
            path_params_submap: if specified, expects the result to have at least theses path parameters, as provided
             by the starlette request
            query_params: If specified, expects the request to have these query arguments exactly
            query_params_submap: If specified, expects the request to have at least these query arguments
            close: If specified, must be a tuple of sender and an integer. Expects the transcript to be closed by that
             sender with that code.
            accepted: If specified, will only match transcripts of accepted connections (if True) or rejected
             connections (if False)
        """
        super().__init__(headers, headers_submap, path, path_params, path_params_submap, query_params,
                         query_params_submap)
        if messages and messages[0] is ...:
            messages = messages[1:]
            self.any_start = True
        else:
            self.any_start = False

        if messages and messages[-1] is ...:
            messages = messages[:-1]
            self.any_end = True
        else:
            self.any_end = False

        if any(m is ... for m in messages):
            raise TypeError('ExpectedWSTranscript can only contain ellipsis at the start and at the end')

        self.expected_messages: Tuple[ExpectedWSMessage] = messages  # type:ignore[assignment]
        self.accepted = accepted
        self.close = close

    def matches(self, recorded: RecordedWSTranscript):
        """
        Checks whether a recorded transcript matches the expectation
        Args:
            recorded: the recorded transcript

        Returns:
            True if the transcript matches, or a mismatch reason otherwise.
        """
        if recorded.close is None:
            raise RuntimeError('the transcript is not yet done')

        reasons = list(self.scope_mismatch_reasons(recorded))

        if self.close is not None and recorded.close != self.close:
            reasons.append(reason_is_ne('close_code', self.close, recorded.close))
        if self.accepted is not None and recorded.accepted != self.accepted:
            reasons.append(reason_is_ne('accepted', self.accepted, recorded.accepted))

        if not self.any_start and not self.any_end and len(recorded) != len(self.expected_messages):
            reasons.append(f'expected exactly {len(self.expected_messages)} messages, found {len(recorded)}')
        if len(recorded) < len(self.expected_messages):
            reasons.append(f'expected at least {len(self.expected_messages)} messages, found only {len(recorded)}')
        if reasons:
            return MismatchReason(', '.join(reasons))
        # in order to account for any_start, any_end, we check every subsequence of the recorded transcript
        # we store a list of candidate indices for the subsequence start
        if not self.any_start:
            # the case of (not any_start and not any_end) is covered by the first condition
            indices: Iterable[int] = (0,)
        elif not self.any_end:
            indices = (len(recorded) - len(self.expected_messages),)
        else:
            indices = range(0, len(recorded) - len(self.expected_messages) + 1)

        whynots = []
        for start_index in indices:
            for i, expected_message in enumerate(self.expected_messages):
                rec = recorded[start_index + i]
                match = expected_message.matches(rec)
                if not match:
                    whynots.append(f'{rec} did not match {expected_message}: {match}')
                    break
            else:
                return True

        return MismatchReason('could not find expected calls' + ''.join('\n\t' + wn for wn in whynots))


class RecordedWSTranscripts(List[RecordedWSTranscript]):
    """
    A list of recorded WS transcripts, in the order that the connections began
    """

    def assert_not_requested(self):
        """
        asserts that no connections were recorded.
        """
        if self:
            raise AssertionError(f'{len(self)} requests were made')

    def assert_requested(self):
        """
        asserts that at least one transcript was recorded.
        """
        if not self:
            raise AssertionError('No requests were made')

    def assert_requested_once(self):
        """
        asserts that exactly one transcript was recorded.
        """
        if not self:
            raise AssertionError('No requests were made')
        if len(self) > 1:
            raise AssertionError(f'{len(self)} requests were made')

    @overload
    def assert_requested_with(self, expected: ExpectedWSTranscript):
        ...

    @overload
    def assert_requested_with(self, *, messages: Sequence[Union[ellipsis, ExpectedWSMessage]] = (...,),  # noqa: F821
                              headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                              headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                              path: Optional[Union[str, Pattern[str]]] = None,
                              path_params: Optional[Mapping[str, Any]] = None,
                              path_params_submap: Optional[Mapping[str, Any]] = None,
                              query_params: Optional[Mapping[str, Collection[str]]] = None,
                              query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                              close: Optional[Tuple[Sender, int]] = None, accepted: Optional[bool] = True):
        ...

    def assert_requested_with(self, expected: Optional[ExpectedWSTranscript] = None, **kwargs):
        """
        Asserts that the latest transcript recorded matches an expected transcript
        Args:
            expected: an expected transcript.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedWSTranscript.
        """
        if expected and kwargs:
            raise TypeError('method can be called with either expected or keyword args, but not both')
        if not expected:
            if not kwargs:
                raise TypeError('either expected or keyword args must be provided')
            expected = ExpectedWSTranscript(**kwargs)

        if not self:
            raise AssertionError('No requests were made')
        match = expected.matches(self[-1])
        if not match:
            raise AssertionError(str(match))

    @overload
    def assert_requested_once_with(self, expected: ExpectedWSTranscript):
        ...

    @overload
    def assert_requested_once_with(self, *,
                                   messages: Sequence[Union[ellipsis, ExpectedWSMessage]] = (...,),  # noqa: F821
                                   headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                                   headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                                   path: Optional[Union[str, Pattern[str]]] = None,
                                   path_params: Optional[Mapping[str, Any]] = None,
                                   path_params_submap: Optional[Mapping[str, Any]] = None,
                                   query_params: Optional[Mapping[str, Collection[str]]] = None,
                                   query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                                   close: Optional[Tuple[Sender, int]] = None, accepted: Optional[bool] = True):
        ...

    def assert_requested_once_with(self, expected: Optional[ExpectedWSTranscript] = None, **kwargs):
        """
        Asserts that there is only one transcript, and that it matches an expected transcript
        Args:
            expected: an expected transcript.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedWSTranscript.
        """
        if expected and kwargs:
            raise TypeError('method can be called with either expected or keyword args, but not both')
        if not expected:
            if not kwargs:
                raise TypeError('either expected or keyword args must be provided')
            expected = ExpectedWSTranscript(**kwargs)

        if not self:
            raise AssertionError('No requests were made')
        if len(self) > 1:
            raise AssertionError(f'{len(self)} requests were made')
        match = expected.matches(self[0])
        if not match:
            raise AssertionError(str(match))

    @overload
    def assert_any_request(self, expected: ExpectedWSTranscript):
        ...

    @overload
    def assert_any_request(self, *, messages: Sequence[Union[ellipsis, ExpectedWSMessage]] = (...,),  # noqa: F821
                           headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
                           headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
                           path: Optional[Union[str, Pattern[str]]] = None,
                           path_params: Optional[Mapping[str, Any]] = None,
                           path_params_submap: Optional[Mapping[str, Any]] = None,
                           query_params: Optional[Mapping[str, Collection[str]]] = None,
                           query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
                           close: Optional[Tuple[Sender, int]] = None, accepted: Optional[bool] = True):
        ...

    def assert_any_request(self, expected: Optional[ExpectedWSTranscript] = None, **kwargs):
        """
        Asserts that the at least one transcript recorded matches an expected transcript
        Args:
            expected: an expected transcript.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedWSTranscript.
        """
        if expected and kwargs:
            raise TypeError('method can be called with either expected or keyword args, but not both')
        if not expected:
            if not kwargs:
                raise TypeError('either expected or keyword args must be provided')
            expected = ExpectedWSTranscript(**kwargs)

        if not self:
            raise AssertionError('No requests were made')
        if any(expected.matches(req) for req in self):
            return
        raise AssertionError('No transcripts match')



from contextlib import closing, contextmanager

from docker import DockerClient


@contextmanager
def docker_client():
    """
    Starts docker client from the environment, with a fallback to default HTTP port
    (for running from within virtual machines)
    """
    try:
        ret = DockerClient.from_env()
        ret.ping()
    except Exception:  # pragma: no cover
        ret = DockerClient(base_url='tcp://localhost:2375')
        ret.ping()
    with closing(ret):
        yield ret



"""
Azure Blob Storage module, for creating container, uploading files to it and downloading files.
"""
from __future__ import annotations

from typing import List, Optional, Sequence, Union

from docker import DockerClient
from docker.models.networks import Network

from yellowbox.containers import create_and_pull, get_ports, short_id
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import RunMixin, SingleContainerService

__all__ = ['BlobStorageService']

BLOB_STORAGE_DEFAULT_PORT = 10000
DEFAULT_ACCOUNT_KEY = "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw=="
DEFAULT_ACCOUNT_NAME = "devstoreaccount1"
STORAGE_URL_FORMAT = "http://127.0.0.1:{port}/{account}"


class _ResourceNotReady(Exception):
    pass


class BlobStorageService(SingleContainerService, RunMixin):
    """
    Starts Azurite, Azure's storage emulator.
    Provides helper functions for preparing the instance for testing.
    TODO: Make account name and key configurable.
    """
    account_name = DEFAULT_ACCOUNT_NAME
    account_key = DEFAULT_ACCOUNT_KEY

    def __init__(self, docker_client: DockerClient,
                 image: str = "mcr.microsoft.com/azure-storage/azurite:latest",
                 **kwargs):
        container = create_and_pull(
            docker_client, image, "azurite-blob --blobHost 0.0.0.0", publish_all_ports=True)
        super().__init__(container, **kwargs)

    def stop(self, signal: Union[str, int] = 'SIGKILL'):
        """
        We override to change the signal.
        """
        super().stop(signal)

    def client_port(self):
        return get_ports(self.container)[BLOB_STORAGE_DEFAULT_PORT]

    @property
    def connection_string(self):
        """Connection string to connect from host to container"""
        return (
            f"DefaultEndpointsProtocol=http;"
            f"AccountName={self.account_name};"
            f"AccountKey={self.account_key};"
            f"BlobEndpoint={self.endpoint_url};")

    @property
    def container_connection_string(self):
        """Connection string to connect across containers in the same network"""
        return (
            f"DefaultEndpointsProtocol=http;"
            f"AccountName={self.account_name};"
            f"AccountKey={self.account_key};"
            f"BlobEndpoint={self.container_endpoint_url};")

    @property
    def endpoint_url(self):
        """URL for the endpoint from docker host"""
        return f'http://localhost:{self.client_port()}/{self.account_name}'

    @property
    def container_endpoint_url(self):
        """URL for the endpoint from another container"""
        return f'http://{short_id(self.container)}:{BLOB_STORAGE_DEFAULT_PORT}/{self.account_name}'

    @property
    def account_credentials(self):
        """Azure credentials dict to connect to the service"""
        return {'account_name': self.account_name, 'account_key': self.account_key}

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()

        def check_ready():
            if b"Azurite Blob service successfully listens on" not in self.container.logs():
                raise _ResourceNotReady

        retry_spec = retry_spec or RetrySpec(attempts=10)

        retry_spec.retry(check_ready, _ResourceNotReady)
        return self

    def connect(self, network: Network, aliases: Optional[List[str]] = None,
                **kwargs) -> Sequence[str]:
        # Make sure the id is in the aliases list. Needed for the container
        # connection string.
        if aliases is not None:
            aliases.append(short_id(self.container))
        return super().connect(network, aliases=aliases, **kwargs)



from typing import IO, Callable, ContextManager, Mapping, Optional, Sequence, TypeVar, Union

from docker import DockerClient
from redis import ConnectionError as RedisConnectionError, Redis

from yellowbox import RunMixin
from yellowbox.containers import create_and_pull, get_ports, upload_file
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import SingleContainerService

__all__ = ['RedisService', 'REDIS_DEFAULT_PORT', 'DEFAULT_RDB_PATH', 'append_state']

REDIS_DEFAULT_PORT = 6379
DEFAULT_RDB_PATH = "/data/dump.rdb"

_T = TypeVar("_T", bound=ContextManager)
RedisPrimitive = Union[str, int, float, bytes]
RedisState = Mapping[str, Union[RedisPrimitive, Mapping[str, RedisPrimitive], Sequence[RedisPrimitive]]]


def append_state(client: Redis, db_state: RedisState):
    for k, v in db_state.items():
        if isinstance(v, Sequence):
            client.rpush(k, *v)
        elif isinstance(v, Mapping):
            client.hset(k, mapping=v)  # type: ignore
        else:
            client.set(k, v)


class RedisService(SingleContainerService, RunMixin):
    def __init__(self, docker_client: DockerClient, image='redis:latest',
                 redis_file: Optional[IO[bytes]] = None, **kwargs):
        container = create_and_pull(docker_client, image, publish_all_ports=True, detach=True)
        self.started = False
        super().__init__(container, **kwargs)

        if redis_file:
            self.set_rdb(redis_file)

    def set_rdb(self, redis_file: IO[bytes]):
        if self.started:
            raise RuntimeError("Server already started. Cannot set RDB.")
        upload_file(self.container, DEFAULT_RDB_PATH, fileobj=redis_file)

    def client_port(self):
        return get_ports(self.container)[REDIS_DEFAULT_PORT]

    def client(self, *, client_cls: Callable[..., _T] = Redis, **kwargs) -> _T:  # type: ignore
        port = self.client_port()
        return client_cls(host='localhost', port=port, **kwargs)

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        client_cm: ContextManager[Redis] = self.client()
        with client_cm as client:
            retry_spec = retry_spec or RetrySpec(attempts=10)
            retry_spec.retry(client.ping, RedisConnectionError)
        self.started = True
        return self

    def reset_state(self):
        client: Redis
        with self.client() as client:
            client.flushall()

    def set_state(self, db_dict: RedisState):
        client_cm: ContextManager[Redis] = self.client()
        with client_cm as client:
            client.flushall()
            append_state(client, db_dict)



from abc import ABCMeta, abstractmethod
from typing import TypeVar

_T = TypeVar("_T")


class YellowService(metaclass=ABCMeta):
    @abstractmethod
    def start(self: _T) -> _T:
        """
        Start the service.

        Returns:
            self, for usage as a context manager.
        """
        return self

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def is_alive(self) -> bool:
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False



from __future__ import annotations

from contextlib import contextmanager
from copy import deepcopy
from threading import Lock, Thread
from time import sleep
from typing import ContextManager, Iterator, Mapping, Optional, Union, overload

from requests import ConnectionError, HTTPError, get
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Route, WebSocketRoute
from uvicorn import Config, Server
from uvicorn.config import LOGGING_CONFIG as uvicorn_logging_config

from yellowbox import YellowService
from yellowbox.extras.webserver.class_endpoint import HTTPEndpointTemplate, WSEndpointTemplate
from yellowbox.extras.webserver.endpoints import (
    HTTP_SIDE_EFFECT, METHODS, WS_SIDE_EFFECT, MockHTTPEndpoint, MockWSEndpoint, http_endpoint, ws_endpoint
)
from yellowbox.extras.webserver.util import mute_uvicorn_log
from yellowbox.retry import RetrySpec
from yellowbox.utils import docker_host_name


class HandlerError(Exception):
    """
    An exception occurred while handling an endpoint in the webserver thread
    """


class WebServer(YellowService):
    """
    An easy-to-modify HTTP and websocket server, wrapping a starlette application
    """
    _PORT_ACCESS_MAX_RETRIES = 100  # the maximum number of attempts to make when accessing a binding port. Each attempt
    # has an interval of 0.01 seconds

    _CLASS_ENDPOINT_TEMPLATES: Mapping[str, Union[HTTPEndpointTemplate, WSEndpointTemplate]] = {}

    def __init__(self, name: str, port: Optional[int] = None, **kwargs):
        """
        Args:
            name: the name of the service
            port: the port to bind to when serving, default will bind to an available port
            **kwargs: forwarded to the uvicorn configuration.
        """
        self._app = Starlette(debug=True)
        self._route_lock = Lock()

        # since the main thread won't catch errors in handlers, this class will store any error raised while handling,
        #  and raise them in the main thread as soon as we can
        self._pending_exception: Optional[Exception] = None

        if 'log_config' not in kwargs:
            log_config = deepcopy(uvicorn_logging_config)
            log_config['formatters']['access']['fmt'] = \
                f'%(levelprefix)s {name} %(client_addr)s - "%(request_line)s" %(status_code)s'
            log_config['formatters']['default']['fmt'] = f'%(levelprefix)s {name} %(message)s'
            kwargs['log_config'] = log_config

        kwargs.setdefault('host', '0.0.0.0')

        self._port = port

        config = Config(self._app, **kwargs, port=self._port)
        self._server = Server(config)
        self._serve_thread = Thread(name=f'{name}_thread', target=self._server.run)

    @property
    def port(self) -> Optional[int]:
        """
        Returns:
            The port the service is bound to, if the service is binding to anything.

        Notes:
            Will only return None if the port was not provided during construction and the service thread is not running
            If the service is starting up, this property will block until the port is binded, or raise an error if
            blocked for longer than 1 second.
        """
        if self._port or not self._serve_thread.is_alive():
            return self._port
        for _ in range(self._PORT_ACCESS_MAX_RETRIES):
            servers = getattr(self._server, 'servers', None)
            if servers:
                sockets = getattr(servers[0], 'sockets', None)
                if sockets:
                    socket = sockets[0]
                    break
            sleep(0.01)
        else:
            raise RuntimeError('timed out when getting binding port')
        self._port = socket.getsockname()[1]
        return self._port

    @overload
    def add_http_endpoint(self, endpoint: MockHTTPEndpoint) -> MockHTTPEndpoint:
        ...

    @overload
    def add_http_endpoint(self, methods: METHODS, rule_string: str, side_effect: HTTP_SIDE_EFFECT, *,
                          auto_read_body: bool = True, forbid_implicit_head_verb: bool = True) -> MockHTTPEndpoint:
        ...

    def add_http_endpoint(self, *args, **kwargs) -> MockHTTPEndpoint:
        """
        Add an http endpoint to the server
        Args:
            *args: either a single mock http endpoint, or parameters forwarded to http_endpoint construct one
            **kwargs: forwarded to http_endpoint to construct an endpoint

        Returns:
            the http endpoint added to the server
        """
        self._raise_from_pending()
        if len(args) == 1 and not kwargs:
            ep, = args
        else:
            ep = http_endpoint(*args, **kwargs)
        if ep.owner is not None:
            raise RuntimeError('an endpoint cannot be added twice')
        with self._route_lock:
            self._app.routes.append(
                ep.route()
            )
        ep.owner = self
        return ep

    def remove_http_endpoint(self, endpoint: MockHTTPEndpoint):
        """
        Remove an http endpoint previously added to the server
        Args:
            endpoint: the endpoint to remove
        """
        self._raise_from_pending()
        if endpoint.owner is not self:
            raise RuntimeError('endpoint is not added to the server')
        with self._route_lock:
            for i, route in enumerate(self._app.router.routes):
                if isinstance(route, Route) and route.endpoint == endpoint.get:
                    break
            else:
                raise RuntimeError('endpoint is not found in the server')
            self._app.router.routes.pop(i)
            endpoint.owner = None

    @overload
    def patch_http_endpoint(self, endpoint: MockHTTPEndpoint) -> ContextManager[MockHTTPEndpoint]:
        ...

    @overload
    def patch_http_endpoint(self, methods: METHODS, rule_string: str, side_effect: HTTP_SIDE_EFFECT, *,
                            auto_read_body: bool = True, forbid_head_verb: bool = True) \
            -> ContextManager[MockHTTPEndpoint]:
        ...

    @contextmanager  # type:ignore[misc]
    def patch_http_endpoint(self, *args, **kwargs) -> Iterator[MockHTTPEndpoint]:
        """
        A context manager to add and then remove an http endpoint
        Args:
            *args: forwarded to self.add_http_endpoint
            **kwargs: forwarded to self.add_http_endpoint

        Returns:
            The temporarily added endpoint
        """
        ep = self.add_http_endpoint(*args, **kwargs)
        try:
            yield ep
        finally:
            self.remove_http_endpoint(ep)

    @overload
    def add_ws_endpoint(self, endpoint: MockWSEndpoint) -> MockWSEndpoint:
        ...

    @overload
    def add_ws_endpoint(self, rule_string: str, side_effect: WS_SIDE_EFFECT) -> MockWSEndpoint:
        ...

    def add_ws_endpoint(self, *args, **kwargs):
        """
        Add a websocket endpoint to the server
        Args:
            *args: either a single mock ws endpoint, or parameters forwarded to ws_endpoint construct one
            **kwargs: forwarded to ws_endpoint to construct an endpoint

        Returns:
            the websocket endpoint added to the server
        """
        self._raise_from_pending()
        if len(args) == 1 and not kwargs:
            ep, = args
        else:
            ep = ws_endpoint(*args, **kwargs)

        if ep.owner is not None:
            raise RuntimeError('an endpoint cannot be added twice')

        with self._route_lock:
            self._app.routes.append(
                WebSocketRoute(ep.rule_string, ep.endpoint, name=ep.__name__)
            )
        ep.owner = self
        return ep

    def remove_ws_endpoint(self, endpoint: MockWSEndpoint):
        """
        Remove a websocket endpoint previously added to the server
        Args:
            endpoint: the endpoint to remove
        """
        self._raise_from_pending()
        if endpoint.owner is not self:
            raise RuntimeError('endpoint is not added to the server')
        with self._route_lock:
            for i, route in enumerate(self._app.router.routes):
                if isinstance(route, WebSocketRoute) and route.app == endpoint.endpoint:
                    break
            else:
                raise RuntimeError('endpoint is not found in the server')
            self._app.router.routes.pop(i)
            endpoint.owner = None

    @overload
    def patch_ws_endpoint(self, endpoint: MockWSEndpoint) -> ContextManager[MockWSEndpoint]:
        ...

    @overload
    def patch_ws_endpoint(self, rule_string: str, side_effect: WS_SIDE_EFFECT) -> ContextManager[MockWSEndpoint]:
        ...

    @contextmanager  # type:ignore[misc]
    def patch_ws_endpoint(self, *args, **kwargs):
        """
        A context manager to add and then remove a ws endpoint
        Args:
            *args: forwarded to self.add_ws_endpoint
            **kwargs: forwarded to self.add_ws_endpoint

        Returns:
            The temporarily added endpoint
        """
        ep = self.add_ws_endpoint(*args, **kwargs)
        try:
            yield ep
        finally:
            self.remove_ws_endpoint(ep)

    def local_url(self, schema: Optional[str] = 'http') -> str:
        """
        Get the url to access this server from the local machine
        Args:
            schema: the optional schema of the url, defaults to http
        """
        if schema is None:
            return f'localhost:{self.port}'
        return f'{schema}://localhost:{self.port}'

    def container_url(self, schema='http') -> str:
        """
        Get the url to access this server from a docker container running in the local machine
        Args:
            schema: the optional schema of the url, defaults to http
        """
        if schema is None:
            return f'{docker_host_name}:{self.port}'
        return f'{schema}://{docker_host_name}:{self.port}'

    def start(self, retry_spec: Optional[RetrySpec] = None) -> WebServer:
        if self._serve_thread.is_alive():
            raise RuntimeError('thread cannot be started twice')
        with mute_uvicorn_log():
            self._serve_thread.start()
            with self.patch_http_endpoint('GET', '/__yellowbox/ping', side_effect=PlainTextResponse('')):
                retry_spec = retry_spec or RetrySpec(interval=0.1, timeout=5)
                retry_spec.retry(
                    lambda: get(self.local_url() + '/__yellowbox/ping').raise_for_status(),
                    (ConnectionError, HTTPError)
                )

        # add all the class endpoints
        for name, template in type(self)._CLASS_ENDPOINT_TEMPLATES.items():
            ep: Union[MockHTTPEndpoint, MockWSEndpoint]
            if isinstance(template, HTTPEndpointTemplate):
                ep = template.construct(self)
                self.add_http_endpoint(ep)
            else:
                assert isinstance(template, WSEndpointTemplate)
                ep = template.construct(self)
                self.add_ws_endpoint(ep)
            setattr(self, name, ep)

        return super().start()

    def stop(self):
        with mute_uvicorn_log():
            self._server.should_exit = True
            self._serve_thread.join()
        super().stop()
        self._raise_from_pending()

    def is_alive(self) -> bool:
        self._raise_from_pending()
        return self._serve_thread.is_alive()

    def _raise_from_pending(self):
        # if there is a pending exception, this will raise it
        if self._pending_exception:
            pending = self._pending_exception
            self._pending_exception = None
            raise HandlerError() from pending

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _cls_endpoints = {}
        for base in cls.__bases__:
            base_http_templates = getattr(base, '_CLASS_ENDPOINT_TEMPLATES', None)
            if base_http_templates:
                overlapping_keys = base_http_templates.keys() & _cls_endpoints.keys()
                if overlapping_keys:
                    raise TypeError(f'overlapping cls endpoints: {overlapping_keys}')
                _cls_endpoints.update(base_http_templates)

        for k, v in vars(cls).items():
            if isinstance(v, (HTTPEndpointTemplate, WSEndpointTemplate)):
                if k in _cls_endpoints:
                    raise TypeError(f'cls endpoint {k} already defined')
                _cls_endpoints[k] = v

        cls._CLASS_ENDPOINT_TEMPLATES = _cls_endpoints



from __future__ import annotations

import io
import os
import shutil
import stat
import tarfile
from contextlib import contextmanager
from os import PathLike
from tempfile import TemporaryFile
from typing import IO, Dict, Generator, List, Sequence, TypeVar, Union, cast

import docker
from docker import DockerClient
from docker.errors import ImageNotFound
from docker.models.containers import Container
from docker.models.networks import Network
from requests import HTTPError

__all__ = ['get_ports', 'get_aliases', 'is_alive', 'is_removed', 'killing', 'create_and_pull',
           'download_file', 'upload_file', 'SafeContainerCreator']

_DEFAULT_TIMEOUT = 10

_T = TypeVar("_T")
_CT = TypeVar("_CT", bound=Container)


def get_ports(container: Container) -> Dict[int, int]:
    """Get the exposed (published) ports of a given container

    Useful for when the ports are assigned dynamically.

    Example:
        >>> c = Container("redis:latest", publish_all_ports=True)
        >>> c.run()
        >>> c.reload()  # Must reload to get updated config.
        >>> ports = get_ports(c)
        >>> ports[6379]  # Random port assigned to 6379 inside the container
        1234

    Note: Container must be up and running. To make sure data is up-to-date,
    make sure you .reload() the container before attempting to fetch the ports.

    Args:
        container: Docker container.

    Returns:
        Port mapping {internal_container_port: external_host_port}.
    """
    # todo this probably won't work for multi-network containers
    ports = {}
    portmap = container.attrs["NetworkSettings"]["Ports"]
    for port, external_address in portmap.items():
        # Filter out unpublished ports.
        if external_address is None:
            continue

        assert len(external_address) > 0

        external_port = int(external_address[0]["HostPort"])

        port, *_ = port.partition("/")  # Strip out type (tcp, udp, ...)
        ports[int(port)] = int(external_port)

    return ports


def get_aliases(container: Container, network: Union[str, Network]) -> Sequence[str]:
    if not isinstance(network, str):
        network = network.name
    return container.attrs["NetworkSettings"]["Networks"][network]["Aliases"]


def short_id(container: Container) -> str:
    """Get the short 12-character id of a container

    By default, the short 12-character id is used as a network alias in all
    of the networks connected to the container.

    Args:
        container: Docker container to retrieve ID from.

    Returns:
        12 character string.
    """
    return container.id[:12]


def is_alive(container: Container) -> bool:
    if is_removed(container):
        return False
    return container.status.lower() not in ('exited', 'stopped')


@contextmanager
def killing(container: _CT, *, timeout: float = _DEFAULT_TIMEOUT,
            signal: str = 'SIGKILL') -> Generator[_CT, None, None]:
    """A context manager that kills a docker container upon completion.

    Example:
        container = DockerContainer(...)
        with killing(container):
            ...
        # Container is now killed with SIGKILL

    Args:
        container: Container to kill upon completion
        timeout: Time to wait for container to be killed after sending a signal.
        Defaults to 10 seconds.
        signal: The signal to send to the container.

    Returns:
        A context manager to be used in a 'with' statement.
    """
    try:
        yield container
    finally:
        if is_alive(container):
            container.kill(signal)
            container.wait(timeout=timeout)


def create_and_pull(docker_client: DockerClient, image: str, *args, **kwargs) -> Container:
    """
    Create a docker container, pulling the image if necessary.
    Args:
        docker_client: the docker client to use.
        image: the image name to create
        *args: additional arguments forwarded to ``docker_client.containers.create``
        **kwargs: additional arguments forwarded to ``docker_client.containers.create``

    Returns:
        A non-started container.

    Note:
        Due to inconsistent behaviour of docker's "pull" command across
        platforms, this function will raise an error if no tag is specified
    """
    name, _, tag = image.partition(":")
    if not tag:
        raise ValueError('the image name must contain a tag')
    try:
        ret = docker_client.containers.create(image, *args, **kwargs)
    except ImageNotFound:
        docker_client.images.pull(image, platform=None)
        ret = docker_client.containers.create(image, *args, **kwargs)
    return ret


def is_removed(container: Container):
    try:
        container.reload()
    except HTTPError:
        return True
    return False


def download_file(container: Container, path: Union[str, PathLike[str]]
                  ) -> IO[bytes]:
    """Download a file from the given container

    Args:
        container: Docker container at any state.
        path: File path.

    Raises:
        FileNotFoundError: Path was not found.
        IsADirectoryError: Path is not a regular file.
    """
    realpath = os.fspath(path)
    exc: Exception
    try:
        iterator, stats = container.get_archive(realpath, chunk_size=None)  # noqa
    except docker.errors.NotFound:
        exc = FileNotFoundError(realpath)
        exc.filename = realpath
        raise exc

    if stat.S_ISDIR(stats["mode"]):
        exc = IsADirectoryError(path)
        exc.filename = realpath
        raise exc

    # Finalizer ensures temporary file will close and be removed.
    temp_file = TemporaryFile("w+b")

    for chunk in iterator:
        temp_file.write(chunk)
    temp_file.seek(0)

    tar_file = tarfile.open(fileobj=temp_file)
    member = tar_file.next()
    return cast('IO[bytes]', tar_file.extractfile(member))


def upload_file(container: Container, path: Union[str, PathLike[str]],
                data: bytes = None, fileobj: IO[bytes] = None) -> None:
    """Upload a file to the given container

    Args:
        container: Docker container.
        path: Path to upload the file to.
        data: Bytes of data to upload. Cannot be set with fileobj.
        fileobj: File object to uplaod. Cannot be set with data.
    """
    if data is fileobj is None:
        raise TypeError("data or fileobj must be set.")

    if data is not None is not fileobj:
        raise TypeError("Can't set both data and fileobj.")

    filename = os.path.basename(path)

    tar_data = _create_tar(filename, data, fileobj)

    container.put_archive(os.path.dirname(path), tar_data)


def _create_tar(filename, data=None, fileobj=None) -> bytes:
    """Create a tarfile made of the given data

    Args:
        filename: Name of the file to create inside the tar
        data: Data of the file. Cannot exist with fileobj.
        fileobj: File object. Cannot exist with data.

    Returns:
        Bytes of a tarfile, containing the given file.
    """
    output = io.BytesIO()
    with tarfile.open(fileobj=output, mode="w") as tar:
        if data is not None:
            tarinfo = tarfile.TarInfo(name=filename)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, io.BytesIO(data))
        else:
            try:
                # Attempt to extract info (such as size) from file object
                tarinfo = tar.gettarinfo(arcname=filename, fileobj=fileobj)
                tar.addfile(tarinfo, fileobj)
            except (OSError, AttributeError):
                # Failed to extract info, writing and reading from temp file.
                with TemporaryFile("w+b") as temp_file:
                    shutil.copyfileobj(fileobj, temp_file)
                    temp_file.seek(0)
                    temp_file.flush()
                    tarinfo = tar.gettarinfo(arcname=filename, fileobj=temp_file)  # noqa
                    tar.addfile(tarinfo, temp_file)
    return output.getvalue()


class SafeContainerCreator:
    """
    A class that can safely pull and create multiple containers in succession, where if one fails, all the previous
     ones are removed
    """
    def __init__(self, client: DockerClient):
        self.client = client
        self.created: List[Container] = []

    def create_and_pull(self, image, command=None, **kwargs):
        try:
            container = create_and_pull(self.client, image, command, **kwargs)
        except Exception:
            for container in reversed(self.created):
                container.remove(v=True)
            raise
        self.created.append(container)
        return container



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yellowbox', 'yellowbox.extras', 'yellowbox.extras.webserver']

package_data = \
{'': ['*']}

install_requires = \
['docker>=4.2.0,<6.0.0', 'requests', 'yaspin>=1.0.0,<2.0.0']

extras_require = \
{'_all': ['redis>=3.3.0',
          'pika',
          'kafka-python',
          'azure-storage-blob>=12.0.0,<12.7.0',
          'cffi>=1.14.0,<2.0.0',
          'sqlalchemy>=1.3.0,<2.0.0',
          'psycopg2>=2.8.6,<3.0.0',
          'simple_websocket_server',
          'python-igraph>=0.9.6,<0.10.0,!=0.9.8',
          'starlette',
          'uvicorn',
          'websockets',
          'hvac'],
 'azure': ['azure-storage-blob>=12.0.0,<12.7.0', 'cffi>=1.14.0,<2.0.0'],
 'kafka': ['kafka-python'],
 'postgresql': ['sqlalchemy>=1.3.0,<2.0.0', 'psycopg2>=2.8.6,<3.0.0'],
 'rabbit': ['pika'],
 'redis': ['redis>=3.3.0'],
 'vault': ['hvac'],
 'webserver': ['python-igraph>=0.9.6,<0.10.0,!=0.9.8',
               'starlette',
               'uvicorn',
               'websockets'],
 'websocket': ['simple_websocket_server']}

setup_kwargs = {
    'name': 'yellowbox',
    'version': '0.6.8',
    'description': '',
    'long_description': '# Yellowbox\n![Test YellowBox](https://github.com/biocatchltd/yellowbox/workflows/Test%20YellowBox/badge.svg?branch=master)\n[![Coverage](https://codecov.io/github/biocatchltd/yellowbox/coverage.svg?branch=master)](https://codecov.io/github/biocatchltd/yellowbox?branch=master)\n\n\nYellowbox makes it easy to run docker containers as part of black box tests.\n## Examples\nSay you want to run a blackbox test on a service that depends on a redis server.\n\n```python\nfrom yellowbox.clients import docker_client\nfrom yellowbox.extras import RedisService\n\n\ndef test_black_box():\n  with docker_client() as docker_client, RedisService.run(docker_client) as redis:\n    redis_port = redis.client_port()  # this the host port the redis\n    ...  # run your black box test here\n  # yellowbox will automatically close the service when exiting the scope\n\n\ndef test_black_box_with_initial_data():\n  # you can use the service\'s built-in utility functions to\n  # easily interoperate the service\n  with docker_client() as docker_client,\n          RedisService.run(docker_client) as redis,\n          redis.client() as client:\n    client.set("foo", "bar")\n  ...\n```\n## Supported Extras\nThe currently supported built-in services are:\n* Kafka: `from yellowbox.extras import KafkaService`\n    * currently, the kafka service supports only plaintext protocol, and always binds to the host port 9092\n* Logstash: `from yellowbox.extras import LogstashService`\n* RabbitMQ: `from yellowbox.extras import RabbitMQService`\n* Redis: `from yellowbox.extras import RedisService`\n\nNote: all these extras require additional dependencies as specified in the project\'s `extras`.\n## Networks\nYellowbox also makes it easy to set up temporary docker networks, so that different containers and services can\ncommunicate directly.\n\n```python\nfrom yellowbox.clients import docker_client\nfrom yellowbox import temp_network, connect\nfrom yellowbox.extras import RabbitMQService\n\n\ndef test_network():\n  with docker_client() as docker_client,\n          RabbitMQService.run(docker_client) as rabbit,\n          temp_network(docker_client) as network,\n          connect(network, rabbit) as alias:\n    # yellow\'s "connect" function connects between a network and a\n    # Container/YellowService, retrieves the container\'s alias(es) on \n    # the network, and disconnects the two when done\n    another_container = docker_client.containers.create("my-image",\n                                                        environment={"RABBITMQ_HOSTNAME": alias[0]}\n                                                        )\n    with connect(network, another_container):\n      another_container.start()\n      another_container.wait()\n```\n## As Pytest Fixtures\nBoth yellow services and networks can be used fluently with `pytest` fixures\n\n```python\nfrom yellowbox.clients import docker_client\nfrom pytest import fixture\n\nfrom yellowbox.extras import RedisService\n\n\n@fixture\ndef docker_client():\n  with docker_client() as ret:\n    yield ret\n\n\n@fixture\ndef redis_service(docker_client):\n  with RedisService.run(docker_client) as service:\n    yield service\n\n\ndef black_box(redis_service):\n  # run your test with the redis service provided\n  ...\n```\nsince docker container may take some time to set up, it\'s advisable to set their scope to at least `"module"`\n## Extending Yellow\nUsers can create their own Yellow Service classes by implementing the `YellowService` abstract class.\nIf the service encapsulates only a single container, the `SingleContainerService` class already implements\nthe necessary methods.\n\n## License\nYellowbox is registered under the MIT public license\n',
    'author': 'biocatch ltd',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/biocatchltd/yellowbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from contextlib import contextmanager
from typing import Any, Iterable, Iterator, List, Mapping, Optional, Tuple

import hvac
from docker import DockerClient
from hvac.exceptions import VaultError
from requests.exceptions import ConnectionError

from yellowbox.containers import create_and_pull, get_ports
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import RunMixin, SingleContainerService
from yellowbox.utils import docker_host_name

__all__ = ['VAULT_DEFAULT_PORT', 'DEV_POLICY', 'VaultService']

VAULT_DEFAULT_PORT = 8200

DEV_POLICY = {
    "path": {
        "secret/*": {
            "capabilities": ["read"]
        },
    }
}


class VaultService(SingleContainerService, RunMixin):
    """
    A yellow service for Hashicorp Vault, with hvac as a client. The vault container is in dev mode AND SHOULD ONLY BE
    USED FOR DEVELOPMENT.
    """

    def __init__(self, docker_client: DockerClient, image='vault:latest', root_token: str = 'guest', **kwargs):
        """
        Args:
            docker_client: the client to use
            image: the image name and tag of the vault image
            root_token: the token string for the new vault container, with root access
            **kwargs: forwarded to SingleContainerService
        """
        container = create_and_pull(docker_client, image, publish_all_ports=True, detach=True, environment={
            'VAULT_DEV_ROOT_TOKEN_ID': root_token,
        })
        self.started = False
        self.root_token = root_token
        super().__init__(container, **kwargs)

    def client_port(self):
        return get_ports(self.container)[VAULT_DEFAULT_PORT]

    def local_url(self):
        return f'http://127.0.0.1:{self.client_port()}'

    def container_url(self):
        return f'http://{docker_host_name}:{self.client_port()}'

    def sibling_container_url(self, container_alias):
        return f'http://{container_alias}:{VAULT_DEFAULT_PORT}'

    @contextmanager
    def client(self, **kwargs) -> Iterator[hvac.Client]:  # type: ignore
        """
        Get a context manager that creates and closes a root-access hvac client
        Args:
            **kwargs: forwarded to hvac.Client constructor
        """
        client = hvac.Client(url=self.local_url(), token=self.root_token, **kwargs)
        try:
            yield client
        finally:
            client.adapter.close()

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()

        retry_spec = retry_spec or RetrySpec(attempts=10)

        def check_health():
            with self.client() as client:
                client.lookup_token()

        retry_spec.retry(check_health, (VaultError, ConnectionError))
        with self.client() as client:
            client.sys.enable_auth_method('userpass')
        self.started = True
        return self

    def set_users(self, userpass: Iterable[Tuple[str, str]], policy_name='dev', policy: Optional[Mapping] = DEV_POLICY):
        """
        Create and set users with policies
        Args:
            userpass: an iterable of users and passwords to create
            policy_name: the name of the policy to assign to the new users
            policy: if assigned, will create or override `profile_name` with the policy dict provided. Default is a
             policy that can read all secrets
        """
        with self.client() as client:
            if policy:
                client.sys.create_or_update_policy(policy_name, policy)
            for username, password in userpass:
                client.auth.userpass.create_or_update_user(username, password, policies=[policy_name])

    def set_secrets(self, secrets: Mapping[str, Mapping[str, Any]]):
        """
        Create and set secrets in the vault
        Args:
            secrets: a dict of secrets, mapping paths to json objects to store
        """
        with self.client() as client:
            for k, v in secrets.items():
                client.secrets.kv.create_or_update_secret(k, v)

    def clear_secrets(self, root_path='/'):
        """
        permanently remove all secrets and subdirectories from a directory in vault
        Args:
            root_path: the root path to delete, default removes all secrets in the vault
        Notes:
            if a secret is assigned to the directory itself, the secret will not be deleted
        """
        if not root_path.endswith('/'):
            raise ValueError('path must end with slash "/"')
        client: hvac.Client
        with self.client() as client:
            def clear_recursive(path):
                secret_names: List[str] = client.secrets.kv.list_secrets(path)['data']['keys']
                for name in secret_names:
                    if name.endswith('/'):
                        # is a directory
                        clear_recursive(path + name)
                    else:
                        # is a secret
                        client.secrets.kv.delete_metadata_and_all_versions(path + name)

            clear_recursive(root_path)



from __future__ import annotations

import json
import logging
import selectors
import socket
import threading
from typing import Any, Callable, Dict, Iterator, List, TypeVar, Union, cast
from weakref import WeakMethod

from yellowbox.subclasses import YellowService
from yellowbox.utils import docker_host_name

__all__ = ['FakeLogstashService']
_logger = logging.getLogger(__name__)

_T = TypeVar("_T")

_STOP_TIMEOUT = 5  # Timeout for stopping the service

# Sent on internal connection to signal closing of background thread
_CLOSE_SENTINEL = b"\0"


def _level_to_int(level: Union[str, int]) -> int:
    """Convert a log level in str or int into an int.

    Args:
        level: Can be a string or int

    Returns:
        level as int.
    """
    if not isinstance(level, int):
        level = logging.getLevelName(level.upper())

    # In rare cases it might be false, but it shouldn't generally happen.
    if not isinstance(level, int):
        raise ValueError(f"Unknown level {level!r}.")

    return level


class FakeLogstashService(YellowService):
    """Implements a fake logging service that closely resembles Logstash.

    Accepts external TCP connections, with logs received in the "json_lines"
    format.

    Attributes:
        records: A list of all records received. You are welcome to modify,
        clear or iterate over this during runtime. New logs will be added in the
        order they were received. Each record is a dict - for possible keys see
        attribute docs.
        port: Dynamic port external service should connect to. FakeLogstashService
        automatically binds a free port during initialization unless chosen
        otherwise.
        local_host: Host to connect to from the local machine (=="localhost")
        container_host: Host to connect to from inside containers.
        encoding: Encoding of the json lines received. Defaults to utf-8 per
        specification.
        delimiter: Delimiter splitting between json objects. Defaults to b'\n'
        per specification.

    Example:
        >>> import socket
        >>> import time
        >>> ls = FakeLogstashService()
        >>> ls.start()
        >>> # Logstash Handler internally creates this connection and sends a log.
        >>> s = socket.create_connection((ls.local_host, ls.port))
        >>> s.sendall(b'{"record": "value", "level": "ERROR"}\\n')
        >>> s.close()
        >>> time.sleep(0.01)  # Wait for service to process message
        >>> ls.stop()
        >>> ls.assert_logs("ERROR")
        >>> assert ls.records[0]["record"] == "value"
    """
    delimiter: bytes = b"\n"
    encoding: str = "utf-8"
    local_host: str = "localhost"
    container_host: str = docker_host_name

    def __init__(self, port: int = 0) -> None:
        """Initialize the service.

        Args:
            port: Port to listen on. By default or if set to 0, port is chosen
            by the OS.
        """

        self.records: List[Dict[str, Any]] = []
        """
        Records generated by Python's numerous Logstash packages have at least the
        following keys in common:
            * @timestamp: str - Log timestamp in ISO-8601 format.
            * @version: str - Logstash format version (always 1)
            * message: str - Log message.
            * host: str - Host sending the message.
            * path: str - Path to the module writing the log.
            * tags: List[str] - ?
            * type: str - "Logstash"
            * level: str - An all upper-case name of the log level
            * logger_name: str - Logger name
            * stack_info: Optional[str] - Formatted stacktrace if one exists.

        More keys may be added by the specific software sending the logs.
        """

        root = socket.socket()
        root.bind(("0.0.0.0", port))
        self._root = root

        # Avoiding a cyclic reference.
        _background = WeakMethod(self._background_thread)  # type: ignore # typeshed bug.
        self._thread = threading.Thread(target=lambda: _background()(),
                                        daemon=True)

        self._selector = selectors.DefaultSelector()

        # Sockets used for signalling shutdown
        self._rshutdown, self._wshutdown = socket.socketpair()

        self.port: int = self._root.getsockname()[1]

    def __del__(self):
        # Will never happen while thread is running.
        try:
            # Order is important.
            self._root.close()
            self._selector.close()
            self._rshutdown.close()
            self._wshutdown.close()
        except AttributeError:  # pragma: no cover  # Hard and pointless to test.
            pass

    def _create_data_callback(self, sock: socket.socket) -> Callable[[], None]:
        """Creates a callback to be called every time the socket receives data.

        Args:
            sock: A connected socket.socket().

        Returns:
            A callable that should be called every time the socket receives data.
        """
        # Data chunks received while waiting for a delimiter.
        partial_chunks = []
        delimiter = self.delimiter
        encoding = self.encoding

        def process_socket_data():
            """Process & parse incoming socket data."""
            data = sock.recv(1024)

            # Socket closed
            if not data:
                self._selector.unregister(sock)
                return

            chunks = data.split(delimiter)

            # Single partial chunk (no delimiter)
            # todo: Cap max chunks length. Not a security issue as we're using
            # todo: Yellowbox solely for testing.
            if len(chunks) == 1:
                partial_chunks.append(chunks[0])
                return

            # Combine all partial chunks with first.
            partial_chunks.append(chunks[0])
            chunks[0] = b"".join(partial_chunks)
            partial_chunks[:] = [chunks.pop()]

            chunk = b""  # Linters are stupid. Required for it to pass.

            # Parse all chunks into records.
            try:
                for chunk in chunks:
                    record_dict = json.loads(chunk.decode(encoding))
                    self.records.append(record_dict)
            except json.JSONDecodeError:
                self._selector.unregister(sock)
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
                _logger.exception("Failed decoding json, closing socket. "
                                  "Data received: %s", chunk)
                return

        return process_socket_data

    def _background_thread(self):
        """Background thread processing incoming connections and data"""
        while True:
            events = self._selector.select(timeout=5)  # For signal handling
            for key, mask in events:
                # Handle closing request
                if key.fileobj is self._rshutdown:
                    assert self._rshutdown.recv(
                        len(_CLOSE_SENTINEL)) == _CLOSE_SENTINEL
                    return

                # Handle new connection
                if key.fileobj is self._root:
                    new_socket, _ = self._root.accept()
                    self._selector.register(new_socket, selectors.EVENT_READ,
                                            self._create_data_callback(new_socket))
                    continue

                # Data received, run callback
                try:
                    key.data()
                except Exception:
                    _logger.exception("Unknown error occurred, closing connection.")
                    sock = cast(socket.socket, key.fileobj)
                    self._selector.unregister(sock)
                    sock.shutdown(socket.SHUT_RDWR)
                    sock.close()

    def start(self) -> FakeLogstashService:
        self._root.listen()
        self._selector.register(self._root, selectors.EVENT_READ)
        self._selector.register(self._rshutdown, selectors.EVENT_READ)
        self._thread.start()
        return super(FakeLogstashService, self).start()

    def stop(self) -> None:
        """Stop the service

        Raises:
            RuntimeError: Failed stopping the service. Operation timed out.
        """
        if not self._thread.is_alive():
            return
        self._wshutdown.send(_CLOSE_SENTINEL)
        self._thread.join(_STOP_TIMEOUT)  # Should almost never timeout
        if self._thread.is_alive():  # pragma: no cover
            raise RuntimeError(f"Failed stopping {self.__class__.__name__}.")

        for key in self._selector.get_map().values():
            sock = cast(socket.socket, key.fileobj)
            sock.close()

        self._selector.close()

    def is_alive(self) -> bool:
        """Check if FakeLogstashService is alive.

        Returns:
            Boolean.
        """
        return self._thread.is_alive()

    def filter_records(self, level: Union[str, int]) -> Iterator[Dict[str, Any]]:
        """Filter records in the given level or above."""
        level = _level_to_int(level)

        return (record for record in self.records if
                logging.getLevelName(record["level"]) >= level)

    def assert_logs(self, level: Union[str, int]):
        """Asserts that log messages were received in the given level or above.

        Resembles unittest.assertLogs.

        Args:
            level: Log level by name or number

        Raises:
            AssertionError: No logs above the given level were received.
        """
        level = _level_to_int(level)

        if not any(self.filter_records(level)):
            raise AssertionError(f"No logs of level {logging.getLevelName(level)} "
                                 f"or above were received.")

    def assert_no_logs(self, level: Union[str, int]):
        """Asserts that no log messages were received in the given level or above.

         Args:
             level: Log level by name or number

         Raises:
             AssertionError: A log above the given level was received.
         """
        level = _level_to_int(level)

        record = next(self.filter_records(level), None)
        if record:
            raise AssertionError(f"A log level {record['level']} was received. "
                                 f"Message: {record['message']}")



from yellowbox.extras.webserver.class_endpoint import class_http_endpoint, class_ws_endpoint
from yellowbox.extras.webserver.endpoints import MockHTTPEndpoint, MockWSEndpoint, http_endpoint, ws_endpoint
from yellowbox.extras.webserver.http_request_capture import ExpectedHTTPRequest
from yellowbox.extras.webserver.util import iter_side_effects
from yellowbox.extras.webserver.webserver import WebServer
from yellowbox.extras.webserver.ws_request_capture import ExpectedWSTranscript, Sender

__all__ = ['WebServer', 'http_endpoint', 'ws_endpoint', 'ExpectedHTTPRequest', 'ExpectedWSTranscript', 'Sender',
           'MockHTTPEndpoint', 'MockWSEndpoint', 'class_ws_endpoint', 'class_http_endpoint', 'iter_side_effects']



from contextlib import contextmanager
from typing import Generator, TypeVar, Union
from uuid import uuid1

from docker import DockerClient
from docker.models.containers import Container
from docker.models.networks import Network

from yellowbox.containers import get_aliases, is_removed
from yellowbox.subclasses import ContainerService

__all__ = ['temp_network', 'anonymous_network', 'connect', 'disconnecting']

_T = TypeVar("_T")
_NT = TypeVar("_NT", bound=Network)


def anonymous_network(client: DockerClient, *args, **kwargs) -> Network:
    name = f"yellowbox-{uuid1()}"
    return client.networks.create(name, *args, **kwargs)


@contextmanager
def temp_network(client: DockerClient, name=None, *args, **kwargs):
    """Context manager for creating a temporary Docker network

    Network will be automatically removed upon context manager completion.

    Example:
        >>> with docker_client() as client, temp_network(client) as network:
        ...   print(network.name)
        ...
        yellowbox-...

    Args:
        client: Docker client to create network with.
        name: Network name. Defaults to a randomly generated name, with the
        prefix "yellowbox".
        *args, **kwargs: Extra arguments for docker network creation.

    Returns:
        Context manager for the newly created network
    """
    if name:
        network = client.networks.create(name, *args, **kwargs)
    else:
        network = anonymous_network(client, *args, **kwargs)
    with disconnecting(network, remove=True) as network:
        yield network


@contextmanager
def connect(network: Network, obj: Union[Container, ContainerService], **kwargs):
    """Temporarily connect a container or yellow service into a network.

    Args:
        network: Docker network.
        obj: Container or YellowService to connect to the network.

    Returns:
        Context manager for handling the connection.
    """
    if isinstance(obj, ContainerService):
        ret = obj.connect(network, **kwargs)  # type: ignore[call-arg]
    else:
        network.connect(obj, **kwargs)
        obj.reload()
        ret = get_aliases(obj, network)
    try:
        yield ret
    finally:
        if isinstance(obj, ContainerService):
            obj.disconnect(network)
        elif not is_removed(obj):
            network.disconnect(obj)


@contextmanager
def disconnecting(network: _NT, *, remove: bool = False) -> Generator[_NT, None, None]:
    """A context manager that disconnects a docker network upon completion.

    Example:
        network = DockerNetwork(...)
        with disconnecting(network):
            ...
        # Network is now disconnected from all containers.

    Args:
        network: Network to disconnect upon completion.
        remove: Whether to remove the network upon completion or not.

    Returns:
        A context manager to be used in a 'with' statement.
    """
    try:
        yield network
    finally:
        network.reload()
        for container in network.containers:
            network.disconnect(container)
        if remove:
            network.remove()
